"""
Analysis functions for pom data
05/09/2018

"""
import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

" Creating dataframes and aggregating population biomarker data per simulation "

def fill_in_grouped_sim_data(grouped_sim_data, results, biomarker,
                             sim_types, sim_type_amp_relationship,
                             sim_name_conversions,
                             how_to_fill, filter_biomarkers_for_outliers=False, threshold=None):
    """"
    Fill in a dataframe with aggregated biomarker data
    """    


    # Check inputs
    assert any([amp_scale_factor == 1 for _,amp_scale_factor in sim_type_amp_relationship.items()]), "No base value of 1 for any of the sim types in sim_type_amp_relationship"
        
    for name in grouped_sim_data.index:
        # To do here: get simulations from the conversion dict, do the appropriate averaging for the how_to_fill method, add that data in
        # Then, get gnav18 from the simulation short name and the amplitudes from the long names using the conversion dict.
        # Add these in too.
        # And we're done
        #names = 
        
        # Get simulations
        simulations = sim_name_conversions[name]
        for simulation, sim_type in simulations.items():
            # Input biomarker data
            unaggregated_data = results.loc[:,(simulation,biomarker)]

            if threshold:
                # Only keep rows that are not below or at threshold
                unaggregated_data = unaggregated_data[unaggregated_data <= threshold]
                
            # filtering of biomarkers for outliers
            if filter_biomarkers_for_outliers:
                if biomarker == "APHalfWidth":
                    # Remove outliers
                    outlier_definition = get_outlier_definition(biomarker) # ms
                    outliers = unaggregated_data >= outlier_definition
                    num_outliers = outliers.sum()
                    mask = ~outliers # Invert boolean series
                    unaggregated_data = unaggregated_data[mask]
                    #if num_outliers > 0:
                        #print("Removed {} AP Half Width outliers > {}".format(num_outliers, outlier_definition))

            if how_to_fill == 'mean':
                data = unaggregated_data.mean()
            elif how_to_fill == 'std':
                data = unaggregated_data.std()
            elif how_to_fill == 'median':
                data = unaggregated_data.median()
            elif how_to_fill == 'mean_fillna':
                data = unaggregated_data.fillna(0).mean()
            elif how_to_fill == 'mean_freq_fillna':
                assert biomarker == 'ISI', "Biomarker for frequency needs to be ISI not {}".format(biomarker) 
                # Convert to frequency, then fill nans with 0s and take mean
                unaggregated_data = 1000.0/unaggregated_data
                data = unaggregated_data.fillna(0).mean() 
            elif how_to_fill == 'mean_freq_dropna':
                assert biomarker == 'ISI'
                # Convert to frequency, then DROP nans and take mean
                unaggregated_data = 1000.0/unaggregated_data
                data = unaggregated_data.dropna().mean() 
            elif isinstance(how_to_fill, int):
                # Model index
                data = unaggregated_data.loc[how_to_fill]
            else:
                raise ValueError("How to fill method: {} not supported.".format(how_to_fill))
            
            grouped_sim_data.at[name,(biomarker, sim_type)] = data

            # Input amplitudes
            amp = get_amplitude(simulation, amp_units='pA', delimiter='_')
            grouped_sim_data.at[name, ('Amp', sim_type)] = amp
            
            # Input scaling factors
            scaling_factors = list(grouped_sim_data['Scaling factors'].columns)
            
            for scaling_factor in scaling_factors:
                # Get scaling factor 
                scaling_factor_value = get_parameter_scaling(simulation, scaling_factor, delimiter='_')
                grouped_sim_data.at[name, ('Scaling factors', scaling_factor)] = scaling_factor_value

def make_grouped_sim_data(pop, biomarker='APFullWidth', agg='mean', filter_outliers=False, scaled_parameters=['GNav18'], threshold=None):
    " Aggregate population biomarker results per simulation to analyse at a per simulation level over the ensemble population. "

    sim_types = ['step', 'ramp'] # First entry in list is the base
    sim_type_as_base = sim_types[0] # Use step as base as ramp has amplitude x10 of step
    sim_type_amp_relationship = {'step':1, 'ramp':10}
    assert sim_type_amp_relationship[sim_type_as_base] == 1


    grouped_sim_data = make_empty_grouped_sim_data( pop=pop,
                                                    biomarker=biomarker,
                                                    filter_outliers=filter_outliers,
                                                    scaled_parameters=scaled_parameters,
                                                    sim_types=sim_types,
                                                    sim_type_as_base=sim_type_as_base,
                                                    sim_type_amp_relationship=sim_type_amp_relationship,
                                                    )

    sim_name_conversions = make_sim_name_conversions (pop.get_simulation_names(),
                                                       sim_types,
                                                       sim_type_amp_relationship,
                                                       sim_type_as_base
                                                      )

    fill_in_grouped_sim_data(grouped_sim_data, pop.results, biomarker,
                                 sim_types, sim_type_amp_relationship,
                                 sim_name_conversions,
                                 how_to_fill=agg,
                                 filter_biomarkers_for_outliers=filter_outliers,
                                 threshold=threshold)

    if biomarker is not 'Firing pattern':
        grouped_sim_data = grouped_sim_data.astype(float)
    
    return grouped_sim_data

def make_empty_grouped_sim_data(pop, biomarker='APFullWidth', filter_outliers=False,
                                scaled_parameters=['GNav18'], sim_types=['step','ramp'],
                                sim_type_as_base='step', sim_type_amp_relationship={'step':1, 'ramp':10},
                                ):
    " Aggregate population biomarker results per simulation to analyse at a per simulation level over the ensemble population. "
    
    arrays =[[biomarker]*len(sim_types)+['Amp']*len(sim_types) + ['Scaling factors']*len(scaled_parameters),sim_types*2 + scaled_parameters] # Build multiarray columns
    columns = pd.MultiIndex.from_arrays(arrays, names=['',''])
    sim_names =  pop.get_simulation_names()
    sim_name_conversions = make_sim_name_conversions(sim_names, sim_types, sim_type_amp_relationship, sim_type_as_base)
    short_sim_names = sorted(list(sim_name_conversions.keys()))

    grouped_sim_data = pd.DataFrame(columns=columns, index=short_sim_names)
    return grouped_sim_data

def make_sim_name_conversions(sim_names, sim_types, sim_type_amp_relationship, sim_type_as_base):
    " Make conversion dict from short sim names to full sim names with sim type "
    sim_name_conversions = {}
    # Get list of sim_names common to step and ramp
    # Complexity: ramp and step amps are different so we will just get the name from step
    for sim_type in sim_types:
        for sim_name in sim_names:
            #  # Ignore rheobase simulations and do base sim type first - important to get oreder right in sim_name_conversions
            if (sim_name not in sim_types) & (sim_type in sim_name):
                short_sim_name, _sim_type = process_sim_name(sim_name, sim_types, sim_type_amp_relationship, amp_units='pA', delimiter='_')
                assert _sim_type == sim_type

                # Build up conversion dict from short name to full names and store sim_type
                if short_sim_name in sim_name_conversions.keys():
                    sim_name_conversions[short_sim_name][sim_name] = sim_type
                else:
                    assert sim_type == sim_type_as_base, "Sim type: {}, name:{}, short_sim_name:{}".format(sim_type, sim_name, short_sim_name)
                    sim_name_conversions[short_sim_name] = {sim_name:sim_type}

    return sim_name_conversions

" Data processing "

def process_firing_pattern_data(firing_pattern_percentages, sim_types=['step','ramp'],
                                sim_type_amp_relationship = {'step':1, 'ramp':10}, scaled_parameters=['GNav18'],
                               ):
    """
    Process simulation names to extract simulation parameters and rename to remove stimulation protocol from name
    Flow:
    1. Convert sim names by removing sim type and storing the conversion between full and shortened names.
    2. Create a formatted dataframe and fill in the firing pattern percentages from the values in the original dataframe.
    3. Extract simulation parameters from the simulation name and add to the formatted dataframe.
    
    TODO: This code shares a lot of code with the functions for aggregating biomarker data. Could refactor into one set of functions
    sharing common code.
    
    
    """
    
    # TODO: Could turn these lines creating sim_names and sim_name_conversions 
    # into a function shared with similar code for biomarkers
    sim_type_as_base = sim_types[0]
    sim_names = firing_pattern_percentages.index.tolist()
    short_sim_names = [] # Simulation names without stimulation protocol
    sim_name_conversions = {} # Conversion between full and short sim names
    
    for sim_type in sim_types:
         for sim_name in sim_names:
            if (sim_name not in sim_types) & (sim_type in sim_name): # Remove rheobase simulations
                short_sim_name, _sim_type = process_sim_name(sim_name, sim_types, sim_type_amp_relationship, amp_units='pA', delimiter='_')
                assert _sim_type == sim_type
                
                # Create conversion between names
                if short_sim_name in sim_name_conversions.keys():
                    sim_name_conversions[short_sim_name][sim_name] = sim_type
                else:
                    assert sim_type == sim_type_as_base, (
                        "Sim type: {}, name:{}, short_sim_name:{} is not the base sim type for the sim type amp relationship.".format(
                            sim_type, sim_name, short_sim_name))
                    sim_name_conversions[short_sim_name] = {sim_name:sim_type}

                if sim_type == sim_type_as_base: # Only add step to sim_names to avoid adding ramp as ramp has different amplitude
                    short_sim_names.append(short_sim_name)
    short_sim_names = sorted(short_sim_names)

    formatted_firing_pattern_data = format_firing_pattern_percentages(
                                                            firing_pattern_percentages, 
                                                            short_sim_names,
                                                            sim_name_conversions, 
                                                            scaled_parameters, 
                                                            sim_types, 
                                                            sim_type_amp_relationship,
                                                           )
    return formatted_firing_pattern_data

             

def format_firing_pattern_percentages(firing_pattern_percentages, short_sim_names, sim_name_conversions, scaled_parameters, sim_types, sim_type_amp_relationship):
    """
    Fill in a dataframe with firing pattern percentages for each simulation
    Equivalent to fill_in_grouped_sim_data() but for firing patterns not single numeric biomarkers.
    
    Copy code from fill_in_grouped_sim_data where needed but:
    1. We don't need a how to fill option as we aggregate by percentages always.
    2. We do need to fill in for all firing patterns, not just one biomarker. 
    """
    
    " Create formatted dataframe with column multiindex "
    
    assert len(scaled_parameters) == 1, "Multiple scaled parameters not supported by format_firing_percentages yet"
    firing_pattern_names = firing_pattern_percentages.columns.tolist()
    sim_type_as_base = sim_types[0] # First sim type in list is the base
    assert sim_type_amp_relationship[sim_type_as_base] == 1

    arrays = [list(np.repeat(firing_pattern_names,len(sim_types))) + ['Amp','Amp', 'Scaling factors'], 
              sim_types*(1+len(firing_pattern_names)) + scaled_parameters] # Build multiarray columns
    columns = pd.MultiIndex.from_arrays(arrays, names=['',''])

    formatted_firing_pattern_data = pd.DataFrame(index=short_sim_names, columns=columns)

    " Fill in firing pattern percentages; get and fill in simulation parameters from simulation names "
    for name in formatted_firing_pattern_data.index:
        simulations = sim_name_conversions[name]
        for simulation, sim_type in simulations.items():
            
            # Fill in firing pattern percentage
            for fp_name in firing_pattern_names:
                formatted_firing_pattern_data.at[name, (fp_name, sim_type)] = firing_pattern_percentages.at[simulation,fp_name]
        
            # Fill in stimulus amplitude
            amp = get_amplitude(simulation, amp_units='pA', delimiter='_')
            formatted_firing_pattern_data.at[name, ('Amp', sim_type)] = amp

            # Fill in parameter scaling
            scaling_factors = list(formatted_firing_pattern_data['Scaling factors'].columns)
            for scaling_factor in scaling_factors:
                scaling_factor_value = get_parameter_scaling(simulation, scaling_factor, delimiter='_')
                formatted_firing_pattern_data.at[name, ('Scaling factors', scaling_factor)] = scaling_factor_value
    
    return formatted_firing_pattern_data

def define_region(pop, firing_pattern_thresholds={}, other_thresholds={}, stim_type='step', verbose=False):
    """
    Use some condition e.g. % of models with a given firing pattern to define a region of simulation space to analyse further.
    """

    import operator
    opcodes  = {'>':operator.gt, '>=':operator.ge, '<':operator.lt, '<=':operator.le}

    firing_pattern_percentages = get_firing_pattern_percentages(pop) # Excludes rheobase simulations
    firing_pattern_percentages = process_firing_pattern_data(firing_pattern_percentages)
    region = {}

    # Check each simulation against all thresholds, if it passes them all then add to defined region
    simulations = firing_pattern_percentages.index # Already excluded rheobase simulations and parameters in process_firing_pattern_data
    count = 0
    for simulation in simulations:
        accept_simulation = True

        # Firing pattern thresholds
        for firing_pattern, threshold_vals in firing_pattern_thresholds.items():
            op = opcodes[threshold_vals[0]]
            threshold = threshold_vals[1]
            fp_percentage = firing_pattern_percentages.loc[simulation,(firing_pattern, stim_type)]
            # Check if threshold is accepted
            if not op(fp_percentage, threshold):
                accept_simulation = False

        # Other thresholds (aggregated biomarkers)
        for threshold, val in other_thresholds.items():
            print('Other thresholds (aggregated biomarkers) not implemented yet.')
            # Get aggregated biomarker data and do the same operation on it as for firing pattern but with biomarker values not percentages

        if accept_simulation:
            count += 1
            amplitude = firing_pattern_percentages.loc[simulation,('Amp', stim_type)]
            scaling_factors = dict(firing_pattern_percentages.loc[simulation, 'Scaling factors'])
            region[simulation] = {'Amp':amplitude}
            for scaling_factor, val in scaling_factors.items():
                region[simulation][scaling_factor] = val

    if verbose:
        print('{} simulations out of {} accepted.'.format(count, len(simulations)))
    return region

def get_biomarker_values_by_simulation(results, biomarker_names, simulation_names):
    """
    Transform pom results dataframe to turn make the simulation name a categorical column rather than a multiindex level
    
    Examples
    1 - simple:
    pop = pom.load('nav18_project_2_1_1.pickle')
    get_biomarker_values_by_simulation(pop.results, biomarker_names='APPeak', simulation_names=['ramp','step'])

    2 - multiple biomarkers, list comprehension for simulation names:
    get_biomarker_values_by_simulation(pop.results, 
                                   biomarker_names=['APPeak', 'Threshold', 'Firing pattern'], 
                                   simulation_names=[name for name in pop.results.columns.levels[0] 
                                                     if name not in ['Parameters', 'step', 'ramp']]
                                  )
    """
    data = results.copy()
    data = results.swaplevel(0,1,axis=1) # Swap so top column level is biomarker not simulation
    data = data.sort_index(level=0,axis=1) # Group under each biomarker  
    data = data.loc[:,(biomarker_names, simulation_names)] # Get only needed biomarker and simulations
    data = data.stack().reset_index().set_index('Model') # Remove simulation from column index and turn it into column
    data = data.rename(columns={'':'Simulation'}) # Name the Simulation column
    return data

def pivot_dataframe_for_heatmap(df, x, y, z, labels):
    """ 
    Pivot a dataframe and drop unwanted components so the the columns are x, the rows are y, and the element values are z. 
    Labels should be axis names ordered in a list [x,y,z].
    """
    df_labels = [x,y,z]
    df = pd.DataFrame()
    
    for df_label, heatmap_label in zip(df_labels,labels):
        df[heatmap_label] = grouped_sim_data[df_label]

    # Pivot
    df = df.pivot(index=labels[1], columns=labels[0], values=labels[2])
    df = df.iloc[::-1] # Reverse index to have origin at 0 0
    return df

def pom_consistency_in_region(pop, region, firing_pattern, stim_type, amp_stim_relationship):
    """
    Plot or find mean and std dev of the distribution of firing pattern consistency in the population, 
    within a defined region. See brown notebook 1 for a sketch of histogram and cumulative consistency plots.
    """
    
    " Get consistency for each model "
    model_consistencies = pd.DataFrame(index=pop.results.index, columns=['Consistency'])
    for model in pop.results.index:
        model_consistency = model_consistency_in_region(model, pop, region, firing_pattern, stim_type, amp_stim_relationship)
        model_consistencies.at[model, 'Consistency'] = model_consistency
        
    pom_consistency = model_consistencies
    return pom_consistency

def model_consistency_in_region(model, pop, region, firing_pattern, stim_type, amp_stim_relationship):
    """
    Calculate consistency percentage for one model for the given firing pattern, population and simulation region
    """
    
    num_simulations = 0 # Count total number of simulations or could calc directly from region and check we don't get any errors with 
              # any simulation in the region
    num_consistent_simulations = 0
    for simulation in region:
        num_simulations += 1
        sim_full_name = short_name_to_full(simulation, stim_type, amp_stim_relationship, delimiter='_', amp_units='pA')
        simulation_firing_pattern = pop.results.at[model, (sim_full_name, 'Firing pattern')]
        # Check if firing pattern is part of simulation firing pattern

        # Check for nans
        if type(simulation_firing_pattern) != list:
            if np.isnan(simulation_firing_pattern):
                continue

        if firing_pattern in simulation_firing_pattern:
            num_consistent_simulations += 1
        
    model_consistency = 100.*(num_consistent_simulations/num_simulations)
    return model_consistency

def assign_subpopulation_from_region(pop, region, criteria, verbose=False):
    """
    Compute required consistencies and assign subpopulations to a population of models based on results from
    a simulation region.
    Inputs:
    pop - a PopulationOfModels class
    region - a list of simulations
    criteria - has the format: {name:[firing pattern, opcode, value]}
    E.g. {'single < 25':['single', '>=', 25]}
    
    """
    import operator
    opcodes  = {'>':operator.gt, '>=':operator.ge, '<':operator.lt, '<=':operator.le}
    
    # Build consistencies for each criteria 
    consistencies = pd.DataFrame(index=pop.results.index)
    for criterion in criteria:
        
        firing_pattern = criteria[criterion][0]
        pom_consistency = pom_consistency_in_region(pop, 
                                            region, 
                                            firing_pattern=firing_pattern, 
                                            stim_type='step', 
                                            amp_stim_relationship={'step':1,'ramp':10})
        consistencies[firing_pattern] = pom_consistency
          
    # Find the models that fulfill all consistency criteria
    models_passing_criteria = pd.DataFrame(index=consistencies.index)
    for criterion, criterion_params in criteria.items():
        firing_pattern = criterion_params[0]
        opcode = criterion_params[1]
        val = criterion_params[2]
        op = opcodes[opcode]     
        consistency = consistencies[firing_pattern]
        models_passing_criteria[criterion] = op(consistency,val)
        
    models_passing_all_criteria = models_passing_criteria.all(axis=1)
    # Filter away models that don't pass all criteria
    subpopulation = pd.DataFrame(index=pop.results.index)
    subpopulation = subpopulation[models_passing_all_criteria]
    
    if verbose:
        print('{} models out of {} in population of models are in the subpopulation'.format(
            len(subpopulation.index), len(pop.results.index))
             )
        
    return subpopulation

def calculate_percentage_of_firing_patterns(firing_patterns):
    """
    Calculate percentage of each firing pattern.
    firing_patterns is a series or list of lists of firing patterns
    """
    assert type(firing_patterns) == pd.Series, "firing_patterns is not a pandas Series"
    firing_patterns = firing_patterns.dropna() # Clean up data

    firing_patterns_in_this_sim = []
    for model_fps in firing_patterns: # Iterate over models
        for fp in model_fps: # Iterate over firing patterns for one model
            firing_patterns_in_this_sim.append(fp)

    set_of_firing_patterns_in_this_sim = set(firing_patterns_in_this_sim)

    # Then iterate through and count each one
    counts = {fp:0 for fp in set_of_firing_patterns_in_this_sim}
    for model_fps in firing_patterns:
        for fp in model_fps:
            counts[fp] += 1 
   
    # Then divide each by number of models
    percentages = {fp:None for fp in set_of_firing_patterns_in_this_sim}
    num_models = len(firing_patterns)
    for fp, val in counts.items():
        percentages[fp] = 100.0*float(val)/num_models
    return percentages

" Data analysis "

def pom_consistency_stats(pom_consistency):
    return {'mean':pom_consistency['Consistency'].mean(), 'std':pom_consistency['Consistency'].std()}

def pom_consistency_hist(pom_consistency):
    plt.hist(pom_consistency)

" Plotting "

def make_3d_plot(data, x, y, z, cutoff, fillna_value, labels, angle=(20,300), zticks=None, title=None):
    """
    Construct a 3d plot 
    """
    df = pd.DataFrame(columns=["X","Y","Z"])
    df['X'] = data[x].copy()
    df['Y'] = data[y].copy()
    df['Z'] = data[z].copy()

    df = df.fillna(value=fillna_value,axis=1)
    df.loc[df['Z'] >= cutoff, 'Z'] = fillna_value * 3 # Change cutoff points to be very low
    #df['Z'][df['Z'] >= cutoff] = fillna_value * 3 # old redundant code setting on slice

    # Make the plot
    fig = plt.figure(figsize=(10,10))
    ax = fig.gca(projection='3d')

    # Title
    if title is not None:
        ax.set_title(title, fontsize=18, verticalalignment='baseline')

    " Make the plot pop "
    surf = ax.plot_trisurf(df['X'], df['Y'], df['Z'], cmap=plt.cm.viridis, linewidth=0.2)
    fig.colorbar(surf, shrink=0.5, aspect=5)

    plt.xlabel(labels[0])
    plt.ylabel(labels[1])
    ax.set_zlabel(labels[2], fontsize=20, rotation=40)

    if zticks != None:
        ax.set_zticks(zticks)

    # Rotate it
    ax.view_init(angle[0], angle[1])
    plt.tight_layout()

    plt.show()
    return fig, ax

def visualise_region(pop, region, stim_type, scaling_factor='GNav18'):
    " Visualise region as scatter plot with one scaling factor and amp "
    firing_pattern_percentages = get_firing_pattern_percentages(pop) #  Excludes rheobase simulations
    firing_pattern_percentages = process_firing_pattern_data(firing_pattern_percentages)
    
    for sim in firing_pattern_percentages.index:
        amp = firing_pattern_percentages.loc[sim, ('Amp', stim_type)]
        sf_val = firing_pattern_percentages.loc[sim, ('Scaling factors', scaling_factor)]
        plt.scatter(sf_val, amp, color='k')
        
    # Now plot region
    for sim, sim_vals in region.items():
        amp = sim_vals['Amp']
        sf_val = sim_vals[scaling_factor]
        plt.scatter(sf_val, amp, color='r')

def pom_consistency_cumulative_plot(pom_consistency, bins=10):
    " See https://stackoverflow.com/questions/15408371/cumulative-distribution-plots-python for code source"
    data = pom_consistency['Consistency']
    values, base = np.histogram(data, bins=bins)
    cumulative = np.cumsum(values)
    plt.plot(base[:-1], cumulative)

def plot_biomarker_boxplots(pop, biomarker, stim_amps, sim_type, 
                            sim_types=['step', 'ramp'], 
                            sim_type_amp_relationship={'step':1., 'ramp':10.},
                            base_sim_type='step',
                            save=False,
                            save_path=None,
                            fixed_scaling_factors={},
                           ):
    " Plot boxplots of biomarkers for constant stim amps varying scaling factor "
    
    sim_names = pop.get_simulation_names()
    sim_name_conversions = make_sim_name_conversions(sim_names, sim_types, sim_type_amp_relationship, base_sim_type)
    for stim_amp in stim_amps:
        
        simulations = get_simulations_with_fixed_stim_amp(pop, stim_amp, sim_type, sim_type_amp_relationship, base_sim_type, fixed_scaling_factors=fixed_scaling_factors)
 
        # Build dataframe with the simulation and biomarker value of each model
        df = pd.DataFrame()
        for sim in simulations:

            biomarker_data = pd.DataFrame(columns=[biomarker])
            biomarker_data[biomarker] = pop.results[(sim, biomarker)]
            # Add the scaling factor of the simulation
            sf = get_parameter_scaling(sim, 'GNav18', delimiter='_')
            biomarker_data['Scaling factor'] = sf 
            df = pd.concat([df, biomarker_data])

        df[biomarker] = pd.to_numeric(df[biomarker])
        fig = plt.figure(figsize=(7,8))
        sns.set_context("paper", font_scale=2.2)
        sns.boxplot(x='Scaling factor', y=biomarker, data=df, whis=np.inf)
        plt.title(f'Stim amp = {stim_amp} pA.')
        #plt.xticks([0, 5, 10, 15, 20],[0, 0.5, 1.0, 1.5, 2.0])
        #plt.ylim(0,10)
        plt.ylabel('num APs')
        plt.xlabel('GNav 1.8 scaling factor')
        
        plt.tight_layout()

        for i, sf in enumerate(sorted(set(df['Scaling factor']))):
            plt.scatter(i, df[df['Scaling factor'] == sf][biomarker].mean(), s=40, c='r', marker='x')
        if save:
            if not save_path:
                save_path = f'boxplot_{biomarker}_{stim_amp}.png'
            fig.savefig(save_path, dpi=300)

def plot_biomarker_scatterplots(pop, biomarkers, stim_amps, sim_type,
                          sim_types=['step', 'ramp'], 
                          sim_type_amp_relationship={'step':1, 'ramp':10},
                          base_sim_type='step',
                          selected_scaling_factors=None,
                          plot_together=False,
                          plot=True,
                         ):
    " Plot scatter plots of two biomarkers for constant stim amps varying scaling factor "
    assert len(biomarkers) == 2
    
    sim_names = pop.get_simulation_names()
    sim_name_conversions = make_sim_name_conversions(sim_names, sim_types, sim_type_amp_relationship, base_sim_type)
    dfs = {}
    for stim_amp in stim_amps:
        
        simulations = get_simulations_with_fixed_stim_amp(pop, stim_amp, sim_type, sim_type_amp_relationship, base_sim_type)
        
        # Build dataframe with the simulation and biomarker value of each model
        df = pd.DataFrame()
        for sim in simulations:

            biomarker_data = pd.DataFrame(columns=biomarkers)
            for biomarker in biomarkers:
                biomarker_data[biomarker] = pop.results[(sim, biomarker)]
            # Add the scaling factor of the simulation
            sf = get_parameter_scaling(sim, 'GNav18', delimiter='_')
            biomarker_data['Scaling factor'] = sf 
            df = pd.concat([df, biomarker_data])
        
        df[biomarkers] = df[biomarkers].apply(pd.to_numeric)
        dfs[stim_amp] = df
        
        if plot:
            plotted = False
            scaling_factors = sorted(set(df['Scaling factor']))
            for sf in scaling_factors:
                if (selected_scaling_factors is not None) & (sf not in selected_scaling_factors):
                    continue
                df_sf = df[df['Scaling factor'] == sf]
                if plot_together & (not plotted):
                    f, ax = plt.subplots(1,figsize=(6,6))
                    ax.set(xlabel=biomarkers[0], ylabel=biomarkers[1])
                    plotted = True
                elif not plot_together:
                    f, ax = plt.subplots(1,figsize=(6,6))
                    ax.set(title=sf, xlabel=biomarkers[0], ylabel=biomarkers[1]) 

                ax.scatter(df_sf[biomarkers[0]], df_sf[biomarkers[1]])
                print(str(df_sf[biomarkers[0]].mean()) + " " + str(df_sf[biomarkers[1]].mean()))
                #ax.set(xlim=(0,100))

                # @TODO add saving
        
    return dfs

def plot_parameters(pop):
# Plot all the parameters in a population in a grid of scatterplots

    parameters = pop.results['Parameters'].copy()
    parameter_ranges = {'GNav17':[0.0,0.4], 'GNav18':[0.,4.0], 'GNav19':[0.,4.],
                'GKdr':[0.,4.], 'GKA':[0.,40.], 'GKM':[0.,4.], 'GH':[0.,2.], 'GKleak':[0., 0.2]}

    for col in parameters:
        parameters[col] /= parameter_ranges[col][1]/2

    dark_grey = np.array([1,1,1])*(32/255)
    sns.set(style="white")
    sns.set_context(context="notebook", font_scale = 3.5,)
    g = sns.pairplot(parameters, plot_kws = {'color': dark_grey}, diag_kws = {'color': dark_grey})
    # Set scatter plot ranges
    g.set(xticks=[0,2]); g.set(yticks=[0,2])
    # Set histogram ranges
    axes = g.axes
    for i in range(len(parameters.columns)):
        axes[i,i].set_xlim(0,2)

            
" Utilities "

def get_non_rheobase_sim_names(pop):
    sims = [sim_name for sim_name in pop.results.columns.levels[0] if sim_name not in ['Parameters', 'ramp', 'step']]
    return sims
    
def get_amp_str_from_sim_name(name, amp_units, delimiter):
        amp_units_with_delimiter = '{}{}'.format(delimiter, amp_units)
        end = name.find(amp_units_with_delimiter)
        start = name[0:end].rfind(delimiter)+1 # Find from the end of the reduced string
        amp_string = name[start:end]
        return amp_string

def get_amplitude(name, amp_units='pA', delimiter='_'):
    # Get the stim amplitude
    amp_string = get_amp_str_from_sim_name(name, amp_units, delimiter)
    try:
        assert float(amp_string) == int(amp_string)
    except ValueError: # If fail to convert string to float
        return np.nan

    amp = int(amp_string)
    return amp
    
def get_parameter_scaling(name, parameter_name, delimiter='_'):
    start = name.find(parameter_name)+len(parameter_name)+len(delimiter)
    end = name[start:].find(delimiter)
    if end == -1:
        parameter_string = name[start:]
    else:
        parameter_string = name[start:start+end]
    parameter_scaling_factor = float(parameter_string)
    return parameter_scaling_factor

def get_all_firing_patterns_in_population(pop):
    """
    Get the set of all individual firing patterns found in a population
    """
    sims = get_non_rheobase_sim_names(pop)
    all_firing_patterns = []
    for sim in sims:
        df = pop.results[sim].copy()
        sim_firing_patterns = df['Firing pattern'].dropna() # Drop nans to avoid error in trying to iterate on a nan below
        
        for model_fps in sim_firing_patterns: 
            for fp in model_fps:
                all_firing_patterns.append(fp)
    
    all_firing_patterns = set(all_firing_patterns)
    return all_firing_patterns

def get_firing_pattern_percentages(pop):
    """
    Get the firing pattern percentages for each simulation and each firing pattern present in the
    population of models provided as input.
    """
    
    # Initialise firing pattern percentages dataframe
    sims = get_non_rheobase_sim_names(pop)
    all_firing_patterns = get_all_firing_patterns_in_population(pop)
    firing_pattern_percentages = pd.DataFrame(index=sims, columns=all_firing_patterns)

    # Fill in percentages
    for sim in sims:
        firing_patterns = pop.results[sim]['Firing pattern'].copy()
        percentages = calculate_percentage_of_firing_patterns(firing_patterns)

        # Save percentage for each firing pattern
        for fp, perc in percentages.items():
            firing_pattern_percentages.at[sim,fp] = perc

    # Here we could fill in all the NaNs to 0
    firing_pattern_percentages = firing_pattern_percentages.fillna(0)
    return firing_pattern_percentages

def short_name_to_full(short_name, stim_type, amp_stim_relationship={'step':1, 'ramp':10}, delimiter='_', amp_units='pA'):
    """
    Convert short name of a simulation without stim type and with amp converted by amp_stim_relationship
    to the full name of the simulation.
    E.g.
    With amp_stim_relationship = {'step':1, 'ramp':10}:
    '2_1_2_400_pA_GNav18_0.1' would become:
    For step - '2_1_2_step_400_pA_GNav18_0.1'
    For ramp - '2_1_2_ramp_4000_pA_GNav18_0.1'
    """
    
    s = short_name

    amp_str = get_amp_str_from_sim_name(s, amp_units, delimiter)
    amp_val = get_amplitude(s, amp_units, delimiter)
    
    parts = s.split(delimiter+amp_str, maxsplit=1) 
    # Need maxsplit to avoid splitting on the scaling factor(s) e.g. for amp=0 and GNav18_0.0
    # Need to add delimiter to avoid splitting on the simulation name (eg simulation_5_10 for amp=0)
    parts[0] += (delimiter + stim_type + delimiter)

    stim_type_amp_str = str(amp_val * int(amp_stim_relationship[stim_type])) # Assume that all amp_stim_relationships are expressed as ints
    parts[0] += stim_type_amp_str

    full_name = ''.join(parts)
    return full_name

def process_sim_name(name, sim_types, sim_type_amp_relationship, amp_units='pA', delimiter='_'):
    """
    Helper function for building an index with each row being a combined step and ramp result.
    Remove reference to sim_types and trailing underscores. Can just give one sim_type if you want to avoid problem
    with different sim_types having different amplitudes and not grouping.
    """
    for sim_type in sim_types:
        if sim_type in name:
            # Remove sim type
            new_name = name.replace(sim_type + '_','')
            
            # Convert amplitude according to sim_type - amp relationship
            # by finding text before the amp unit and converting
            
            # Find amplitude
            amp_string = get_amplitude(new_name, amp_units, delimiter)
            amp = float(amp_string)
            
            # Convert amplitude by relationship
            amp_scale_down = sim_type_amp_relationship[sim_type]
            # Warning: this may break if the scale down is something that isn't a factor of the amplitude
            new_amp = amp/amp_scale_down
            assert int(new_amp) == float(new_amp), "Something going wrong with amp: {}, check process_sim_name().".format(amp)
            new_amp_string = str(int(new_amp)) # Convert to int since float == int 
            
            # Put the new_amp_string back into the new name
            new_name = new_name.replace('{}{}{}'.format(amp_string, delimiter, amp_units),
                                        '{}{}{}'.format(new_amp_string, delimiter, amp_units))
            return new_name, sim_type 
        
    return None

def get_simulations_with_fixed_stim_amp(pop, stim_amp, sim_type='step', sim_type_amp_relationship={'step':1., 'ramp':10}, base_sim_type='step', sim_types=['step','ramp'], fixed_scaling_factors={}):
    " Get names of all simulations in population with a given stimulus amplitude and type "
    
    sim_names = pop.get_simulation_names()
    sim_name_conversions = make_sim_name_conversions(sim_names, sim_types, sim_type_amp_relationship, base_sim_type)
    grouped_sim_data = make_empty_grouped_sim_data(pop, biomarker='')

    # Add metadata
    scaling_factors = list(grouped_sim_data['Scaling factors'].columns)
    for name in grouped_sim_data.index:

        simulations = sim_name_conversions[name]
        for simulation, _stim_type in simulations.items():

            amp = get_amplitude(simulation, amp_units='pA', delimiter='_')
            grouped_sim_data.at[name, ('Amp', _stim_type)] = amp

            for scaling_factor in scaling_factors:
                scaling_factor_value = get_parameter_scaling(simulation, scaling_factor, delimiter='_')
                grouped_sim_data.at[name, ('Scaling factors', scaling_factor)] = scaling_factor_value


    # Filter the dataframe to find the simulations we want
    simulations = list(grouped_sim_data[grouped_sim_data[('Amp', sim_type)] == stim_amp].index)
    simulations = [short_name_to_full(sim, sim_type, amp_stim_relationship=sim_type_amp_relationship) for sim in simulations]

    # Filter simulations to only include those with particular values of a scaling factor, if required
    for fixed_sf, val in fixed_scaling_factors.items():
        simulations = [sim for sim in simulations if get_parameter_scaling(sim, fixed_sf, delimiter='_') == val] 

    return simulations

" Constants "


def get_outlier_definition(biomarker):
    """
    Centralised definitions of biomarker outliers for filtering
    """

    if biomarker == "APHalfWidth":
        outlier_definition = 100.0 # ms
    else:
        raise ValueError(f"Biomarker {biomarker} not found.")
    return outlier_definition
            
def get_default_stim_type_amp_relationship(): 
    sim_type_amp_relationship={'step':1, 'ramp':10}
    return sim_type_amp_relationship



# -*- coding: utf-8 -*-
"""
loadneuron.py
Functions to load Neuron properly.
Created on Fri Oct 28 12:12:50 2016 @author: Oliver Britton
"""
import os
import sys
from neuron import h

if 'mechanisms_loaded' not in locals():
    mechanisms_loaded = False

# Get platform
if sys.platform == 'win32':
    platform = 'windows'
elif sys.platform == 'linux':
    platform = 'linux'
else:
    raise Exception('Platform {} not supported.'.format(sys.platform))
    
    
def load_neuron_mechanisms(nrnmech_path=None, verbose=False,):
    """
    Main loading function, loads neuron mechanisms, iff they are not already loaded.
    Can only call nrn_load_dll once, calling it a second time causes a crash, so need to check
    we have loaded mechanisms before calling.
    """
    
    # Get paths
    if nrnmech_path == None:
        nrnmech_path = os.path.join(get_mechanism_dir(), get_mech_lib_name(platform))
    else:
        # If path is provided, check it contains library name
        if get_mech_lib_name(platform) not in nrnmech_path:
            nrnmech_path =  os.path.join(nrnmech_path, get_mech_lib_name(platform))
        else:
            pass
    nrnmech_dir = nrnmech_path.rsplit(os.path.sep,1)[0]
    
    # Load mechanisms if needed
    if are_mechanisms_loaded() == False:
        print('Loading {} from {}.'.format(get_mech_lib_name(platform), nrnmech_dir))
        h.nrn_load_dll(nrnmech_path) # Neuron mechanism loading function

    # If needed, check if mechanisms are loaded and report
    if verbose:
        if are_mechanisms_loaded():
           print("Ion channel models are loaded.")
        else:
            print(f"Ion channel models were not able to be loaded from {nrnmech_path}! Have you compiled the model files in the {nrnmech_dir} directory? See https://www.neuron.yale.edu/neuron/static/docs/nmodl/mswin.html or https://www.neuron.yale.edu/neuron/static/docs/nmodl/unix.html for details.")

def get_mech_lib_name(platform):
    if platform == 'windows':
        mech_lib_name = 'nrnmech.dll'
    elif platform == 'linux':
        mech_lib_name = 'libnrnmech.so'
    else:
        raise Exception('Platform {} not supported.'.format(sys.platform))
    return mech_lib_name



def get_mechanism_dir():
    """
    Returns the models directory the compiled nrnmech.dll is stored in. 
    """
    cur_path = (os.path.dirname(__file__))
    mechanisms_path = os.path.abspath(os.path.join(cur_path, os.path.pardir, 'models'))
    return mechanisms_path


def check_num_name_mechs():
    """
    Example function for checking number and names of mechanisms.
    Reminds me how to do string refs in hoc and get names of mechanisms.
    """
    from neuron import h
    mt = h.MechanismType(0)
    sref = h.ref('')
    for i in range(int(mt.count())):
        mt.select(i)
        mt.selected(sref)
        print(sref[0])
    return mt.count()
    
    
def are_mechanisms_loaded(verbose=False):
    """
    Checks if mechanisms have been loaded. Assumes a certain set of mechanisms are loaded by default
    and that if a dll has been loaded it will contain other mechanisms. 
    If an empty dll is loaded this won't think it's been loaded and could cause a crash.
    """
    
    " Get names of mechanisms that are always loaded "
    default_mechanisms = default_mech_list() 
    
    " Get names of currently loaded mechanims "
    current_mechanisms = h.MechanismType(0)
    str_ref = h.ref('') # hoc string reference
    current_mechanism_names = []
    for i in range(int(current_mechanisms.count())):
        current_mechanisms.select(i)
        current_mechanisms.selected(str_ref)
        current_mechanism_names.append(str_ref[0])
    
    " Get only non-default mechanisms "
    non_default_mechanisms = [mech for mech in current_mechanism_names if mech not in default_mechanisms]
        
    " If non_default_mechanisms is empty we haven't loaded (or we've loaded no new mechanisms!) "
    if not non_default_mechanisms :
        # List is empty
        loaded = False
        if verbose:
            print("No non-default mechanisms detected, ready to load nrnmech.dll.")
    else:
        loaded = True
    return loaded    

    
def get_mechanism_list():
    """ Return list of loaded mechanism names """
    current_mechanisms = h.MechanismType(0)
    str_ref = h.ref('') # hoc string reference
    current_mechanism_names = []
    for i in range(int(current_mechanisms.count())):
        current_mechanisms.select(i)
        current_mechanisms.selected(str_ref)
        current_mechanism_names.append(str_ref[0])
    return  current_mechanism_names


def default_mech_list():
    """ Returns list of default mechanism names included in NEURON. """
    default_mechanisms = ['morphology',
                          'capacitance',
                          'pas',
                          'extracellular',
                          'fastpas',
                          'na_ion',
                          'k_ion',
                          'hh',
                         ]
    return default_mechanisms
    
    
" -- Tests --- "
def test_load_mechanisms():
    """ Load multiple times, if we don't crash, we're good! """    
    load_neuron_mechanisms()
    load_neuron_mechanisms()
    load_neuron_mechanisms()
    print('We didn\'t crash!')




import os

try:
	from setuptools import setup
except ImportError:
	from distutils.core import setup
	
config = {
	'description': 'Population of models simulations for dorsal root ganglion neurons',
	'author': 'Oliver Britton',
	'url': 'https://github.com/oliverbritton/drg-pom',
	'download_url': 'https://github.com/oliverbritton/drg-pom/archive/master.zip',
	'author_email': 'oliverjbritton@gmail.com',
	'version': '1.0.1',
	'install_requires': [
            'nose2',
            'numpy<1.16.3',
            'pandas',
            'pyDOE',
            'matplotlib',
            'seaborn',
            ],
	'packages': ['drgpom', 'drgpom.methods', 'drgpom.methods.biomarkers',
            'drgpom.examples', 'drgpom.examples.data', 'drgpom.models'],
	'scripts': [],
	'name': 'drgpom',
        'include_package_data': True,
        'package_data': {'drgpom':[os.path.join('examples','data', '*.*'),
                                   os.path.join('examples','results','*.*'),
                                   os.path.join('models','*.mod')]
                        }
        
}

setup(**config)



# neuron_biomarkers.py
# calculation of AP biomarkers from neuronal voltage traces

import sys
import numpy as np
import pandas as pd
from scipy import optimize
from matplotlib import pyplot as plt

from . import davidson_biomarkers as db
from .. import simulation_helpers as sh
from .. import analysis as an

# Biomarkers to manage and analyse neuronal simulation data and potentially experimental
# data too

RHEO_FAIL = np.nan # Code to return if rheobase calculation fails.
# np.nan == np.nan returns False so use is not instead. pom code
# relies on this value being nan to interface with pandas correctly.

def calculate_biomarkers(traces, model):
    " Calculate every biomarker and output to dict "
    " TODO: Use the rheobase to work out what simulation to run to calculate biomarkers "
    " off of (at rheobase) "
    # biomarker_names = ['APFullWidth', 'APPeak', 'APRiseTime', 'APSlopeMin', 'APSlopeMax',. 'AHPAmp', 'AHPTau', 'ISI', 'RMP', 'Rheobase']

    biomarkers = calculate_simple_biomarkers(traces, model)
    biomarkers['RMP'] =  np.mean(calculate_rmp(traces))
    # Need to do rheobase separately
    biomarkers['Rheobase'] =  calculate_rheobase(model, amp_step=0.1, amp_max=5, make_plot=False,)
    
    return biomarkers
    
def average_biomarker_values(biomarkers, how_to_handle_nans='return'):
    " Average biomarker values for multiple APs while handling biomarkers that "
    if how_to_handle_nans == 'return': # Advantage is we return nan if there are any nans - good for calibration and trouble shooting - shows up weird models easily. 
        pass
    elif how_to_handle_nans == 'remove': # Risky option. Advantage is we still get a number back in mixed cases of nan and non-nan biomarkers, which is potentially risky as it hides a problem in one or more APs.
        biomarkers = np.array(biomarkers)
        if biomarkers[~np.isnan(biomarkers)].size == 0:
            return np.nan
        else:
            biomarkers = biomarkers[~np.isnan(biomarkers)]
    else:
        raise ValueError("{} is not an accepted nan handling method.".format(how_to_handle_nans))
        
    mean_result = np.mean(biomarkers)
    return mean_result
    
def calculate_simple_biomarkers(traces, model="Not needed", how_to_handle_nans='return'):
    """ Calculate every biomarker that can be calculated from a normal simulation trace and output to dict - rheobase and RMP need to be calculated separately."""
    biomarkers = {}
    # biomarker_names = ['APFullWidth', 'APPeak', 'APRiseTime', 'APSlopeMin', 'APSlopeMax',. 'AHPAmp', 'AHPTau', 'ISI', 'RMP', 'Rheobase']
    def error_handle(filename, traces): # Error handler for finding out why biomarkers are throwing errors.
        import pickle
        print(sys.exc_info())
        print(traces['numAPs'])
        plt.figure()
        for t,v in zip(traces['t'],traces['v']):
            plt.plot(t,v)
        with open(filename, 'wb') as handle:
            pickle.dump(traces, handle)
        print("Error, traces dumped to {}.".format(filename))

    try:
        biomarkers['APFullWidth'] = average_biomarker_values(calculate_ap_full_width(traces,threshold=5.,method='gradient'), how_to_handle_nans)
    except:
        error_handle('fullwidth.pickle',traces)
    
    try:
        biomarkers['APHalfWidth'] = average_biomarker_values(calculate_ap_half_width(traces,threshold=5.,method='gradient'), how_to_handle_nans)
    except:
        error_handle('halfwidth.pickle',traces)
                   
    biomarkers['APPeak'] = average_biomarker_values(calculate_ap_peak(traces),how_to_handle_nans)
    
    try:
        biomarkers['APRiseTime'] = average_biomarker_values(calculate_ap_rise_time(traces,dvdtthreshold=5),how_to_handle_nans)
    except:
        error_handle('risetime.pickle',traces)
    ap_slope_mins, ap_slope_maxs = calculate_ap_slope_min_max(traces)
    biomarkers['APSlopeMin'] = average_biomarker_values(ap_slope_mins, how_to_handle_nans)
    biomarkers['APSlopeMax'] = average_biomarker_values(ap_slope_maxs, how_to_handle_nans)
    biomarkers['Threshold'] = average_biomarker_values(calculate_threshold(traces), how_to_handle_nans)
    
    amp, tau, trough = fit_afterhyperpolarization(traces=traces,dvdt_threshold=5, ahp_model='single_exp', full_output=False)
    """
    try:
        amp, tau = fit_afterhyperpolarization(traces=traces,dvdt_threshold=5, ahp_model='single_exp', full_output=False)
    except:
        error_handle('fitahp.pickle',traces)
        amp=0
        tau=0
    """
    biomarkers['AHPAmp'] =  amp
    biomarkers['AHPTau'] =  tau
    biomarkers['AHPTrough'] = trough
    biomarkers['ISI'] = inter_spike_interval(traces)
    biomarkers['numAPs'] = traces['numAPs']
        
    return biomarkers
    
def compute_model_biomarkers(model=None, mechanisms=None, make_plot=False, sim_kwargs=None, xlims=None):
    " Find all standard biomarkers of a model or mechanism set. "

    biomarkers = {}

    if model == None:
        model = sh.build_model(mechanisms)
    # Else use model
        
    if sim_kwargs:
        sim_kwargs['model'] = model
    else:
        sim_kwargs = sh.get_default_simulation_kwargs(model=model)

    rheobase = calculate_rheobase(model, amp_step=0.1, amp_max=5, make_plot=False, sim_kwargs = sim_kwargs)

    if (isinstance(rheobase,float) == False) & (isinstance(rheobase,int) == False):
        # Rheobase not found, don't calculate other biomarkers
        find_other_biomarkers = False
    else:
        find_other_biomarkers = True

    if sim_kwargs:
        sim_kwargs['amp'] = rheobase
        sim_kwargs['model'] = model
    else: 
        sim_kwargs = sh.get_default_simulation_kwargs(amp=rheobase, model=model)
    sim_kwargs['make_plot'] = make_plot

    if find_other_biomarkers:
        output = sh.simulation(**sim_kwargs) 
        t = output['t']; v = output['v']
        t = t[::2]; v = v[::2] # 20 kHz
        traces = split_trace_into_aps(t,v)
        biomarkers = calculate_simple_biomarkers(traces,model,how_to_handle_nans='remove')

    # RMP
    rmp_kwargs = {'amp':0.0, 'dur':3000., 'delay':0., 'interval':0., 'num_stims':1, 't_stop':3000.}
    for kwarg in sim_kwargs:
        # Write in sim_kwargs where they are not already present in rmp_kwargs
        # so that non-RMP specific kwargs are consistent between simulations
        if kwarg not in rmp_kwargs:
            rmp_kwargs[kwarg] = sim_kwargs[kwarg]

    output = sh.simulation(**rmp_kwargs)
    rmp_t = output['t']; rmp_v = output['v']
    rmp_t = rmp_t[::2]; rmp_v = rmp_v[::2] # 20 kHz
    rmp_traces = split_trace_into_aps(rmp_t,rmp_v)
    rmp = np.mean(calculate_rmp(rmp_traces))

    if (make_plot & (xlims != None)):
        plt.xlim(xlims[0], xlims[1])

    # If we calculated other biomarkers, add extras calculated in separate simulations.
    # If we didn't add to empty dictionary, will leave nans when added to master dataframe
    # which is what we want.
    biomarkers['Rheobase'] = rheobase
    biomarkers['RMP'] = rmp
    return biomarkers
    
" --- Calculation and trace manipulation functions -- "

def split_trace_into_aps(t,v,threshold=0,time_threshold=5, check_voltage_gradient=True):#
    """
    Threshold is at 0 mV which can let RF to cause spurious AP detection unless
    we perform a voltage gradient check, which defaults to True.
    
    -- Old ideas to solve the spurious AP detection problem with threshold at 0 mV --
    One idea is to do split trace and then calculate AP width using a voltage threshold of something like -25 mV.
    Then, if AP width is really long (> 100 ms?), redo the calculation with a lower threshold (0 mV?). 
    If mean AP width is then < 100 ms, we use the new split. We could write a log file to say that this happened,
    with the trace in it. 

    However, that is complex and may break if something comes up I haven't thought of.
    Instead we could reset the default threshold to 0 mV but add in a gradient check on the voltage crossing from below.
    Currently a gradient threshold of 1 mV/ms seems like it should be effective although I don't have any examples of slow
    calcium initated APs to test against. 
    """
	
    # Units for defaults
    # t, time_threshold - ms
    # v, threshold - mV

    assert len(t) == len(v), "v and t length mismatch"

    crossings = []
    time_crossings = np.array([])
    
    # Looks for crossings from below
    for i,voltage in enumerate(v[:-1]):
        if (voltage < threshold) & (v[i+1] >= threshold):
            # Check local voltage gradient if neeeded, if gradient is too small ignore the crossing
            # Time window set to 1.0 to try to counteract bug with averaging too much of the pre-upstroke. 
            if (check_voltage_gradient) & (is_voltage_gradient_too_small(i, t, v, dvdt_threshold=1.0, time_window=1.0)):
                continue # Don't add the crossing if the local voltage gradient is small and we're checking for that
            crossings.append(i)
            time_crossings = np.append(time_crossings,t[i])
                
    # For each crossing, remove all instances within the time threshold, leaving only the first crossing of the threshold 
    grouped_crossings = np.zeros(np.size(crossings),float) 
    
    for i in range(len(crossings)-1):
        if grouped_crossings[i] == 0:
            nearby_crossings = np.array( (time_crossings[i+1:] - time_crossings[i]) < time_threshold )
            # Assign 
            grouped_crossings[i+1:] += nearby_crossings
            assert all(grouped_crossings < 2), "Grouped crossing grouped more than once"
            
    firstCrossIndices = np.where(grouped_crossings == 0)
    # Need to turn crossings into a numpy array to index it with np.where
    firstCrossings = np.array(crossings)[firstCrossIndices]
    numAPs = len(firstCrossings)
    assert numAPs >= 0, "Negative number of APs!"
    
    # Assign time and voltage to traces
    times = []
    voltages = []
    # If 1 or 0 APs, return 1 trace, otherwise...
#    if (numAPs == 0) | (numAPs == 1):
#        times.append(t)
#        voltages.append(v)

    
    """ 
    There are some commented assumptions about where traces begin and end here. The core idea is that all data points in the trace have to be assigned to 1 and only 1 AP. If areas of quiescence are a problem for    particular analysis methods, they will be stripped out by other specialised functions. 
    Our goal in this function is to divide up the trace without leaving any of it out, so that we have everything for any future analysis.
    """
    # If we have multiple APs, for each AP find the minimum value
    # of Vm before the next AP
    if numAPs > 0:
        startIdx = np.zeros(numAPs,int)
        endIdx = np.zeros(numAPs,int)    
        for AP in range(numAPs):
            if AP == 0:
                startIdx[0] = 0 # Start of first AP is beginning of trace
            else:
                startIdx[AP] = endIdx[AP-1]+1 # Start of all other APs is after last AP
        
            if AP == numAPs-1:
                endIdx[AP] = len(v)-1 # End of last AP is end of trace
            else:
                # Calculate end of this trace - end is minimum voltage of this trace
                # From threshold of this AP to just before beginning of next threshold
                voltageDuringCurrentAP = v[firstCrossings[AP]:firstCrossings[AP+1]]
                
                # Get index of minimum voltage AFTER the peak
                max_idx = np.argmax(voltageDuringCurrentAP)
                minVmIdx = np.argmin(voltageDuringCurrentAP[max_idx:])
                endIdx[AP] = firstCrossings[AP] + max_idx + minVmIdx # Don't think I need to minus 1 because Python indices start at 0
                
            times.append(t[startIdx[AP]:endIdx[AP]+1])
            voltages.append(v[startIdx[AP]:endIdx[AP]+1]) # Add 1 to as Python slicing ends 1 before last index
    
        for i in range(len(startIdx)-1):
            assert endIdx[i]+1 == startIdx[i+1], "startIdx and endIdx don't match up."
    # Case for no APs - numAPs causes problems here so set indices manually
    elif numAPs == 0:
        times.append(t)
        voltages.append(v)
        startIdx = np.array([0],int)
        endIdx = np.array([len(v)-1],int)
        
    
    assert startIdx[0] == 0, "First AP doesn't start at beginning of trace."
    assert endIdx[-1] == len(v)-1, "Last AP doesn't end at end of trace."
    
    return{'t':times, 'v':voltages, 'startIndices':startIdx, 'endIndices':endIdx, 'numAPs':numAPs}   
SplitTraceIntoAPs = split_trace_into_aps # Alias
    
def voltage_gradient(t,v, method='gradient'):
    # There is a gradient function in numpy to take central differences
    if method == 'gradient':
        dvdt = np.gradient(v,t)# Central differences except at end points
    elif method == 'diff':
        dvdt = np.diff(v)/np.diff(t) # Difference between adjacent points 
    else :
        raise ValueError("Method not found.")
    return dvdt
VoltageGradient = voltage_gradient # Alias


def is_voltage_gradient_too_small(i, t, v, dvdt_threshold, time_window):
    """
    Check if the voltage gradient around the threshold crossing from below at v[i] to v[i+1]
    is too small to have a reasonable likelihood of being a real AP.
    Inputs:
    i - index at which threshold is crossed, between v[i] and v[i+1]
    t,v - time and voltage arrays
    dvdt threshold - mV/ms
    time window (either side of indices i and i+1, so effective window is double the size) - ms
    """ 
    voltage_gradient_too_small = False
    
    # Get time window around v[i] and v[i+1]
    lower_t_bound = t[i] - time_window
    if lower_t_bound < 0: lower_t_bound = 0
    upper_t_bound = t[i+1] + time_window

    # Get indices of t and v that are within the window
    t = np.array(t)
    window_indices = (t >= lower_t_bound) & (t <= upper_t_bound)
    _t = t[window_indices]

    _v = v[window_indices]
    _dvdt = np.gradient(_v,_t)

    # Check mean gradient against threshold
    if np.mean(_dvdt) < dvdt_threshold:
        voltage_gradient_too_small = True

    return voltage_gradient_too_small

    
# --- Biomarkers ---
def rmp(v):
    # RMP should be calculated from a quiescent trace (no stimulus)
    # Ignore first 90% of trace to remove artifacts
    vLen = len(v)
    startIdx = 90*vLen//100
    RMP = min(v[startIdx:])
    RMPIdx = np.argmin(v[startIdx:]) 
    return RMP, RMPIdx
RMP = rmp # Alias
    
def input_res(t, v, current_injection_time):
    # Input resistance calculated from a protocol with an equilibration phase
    # to get to RMP, followed by a sustained small current input.
    # Input res then = (v[-1] - v[RMP])/(I-0)
    # In Davidson, 50 to 100 pA current pulse was used to determine input resistance.
    
    # Divide trace at injection time:
    
    
    # Get RMP at t < injection time:
    
    # Get RMP at t > injection time:
    
# Rheobase - find the first trace with an action potential
# Assumes traces are sorted in order from smallest amplitude upwards
    """ To Do - check that simulations is a bunch of simulations, not just an array """
def rheobase(simulations,amps):
    # Check amps is sorted
    for i in range(len(amps)-1):
        assert amps[i+1] > amps[i], 'Amps in rheobase biomarker not increasing monotonically!'
        
    for simulation,amp in zip(simulations,amps):
        # Search for first trace which produces an AP        
        result = SplitTraceIntoAPs(simulation['t'],simulation['v'])
        if result['numAPs'] > 0:
            return {'rheobase':amp, 'trace':simulation}       
    # If no APs found
    return {'rheobase':np.nan, 'trace':[]}
Rheobase = rheobase # Alias
    
def ap_peak(v):

    peak = max(v)
    location = np.argmax(v)
    return [peak,location]
APPeak = ap_peak # Alias
    
def threshold(t, v, dvdt_threshold=5.):
    # Calculation of threshold voltage as described in Davidson et al., 2014 PAIN using gradient
    # Threshold is in V/s - default of 5 is what was used by Davidson et al.
    dvdt = np.gradient(v,t)    
    thresholds = []
    for i, gradient in enumerate(dvdt[0:-1]):
        if (gradient < dvdt_threshold) & (dvdt[i+1] > dvdt_threshold): # Look for crossing of threshold
            thresholds.append(v[i])
    if thresholds:
        return thresholds[0] # Only use first threshold of crossing
    else:
        return np.nan
        
def ap_rise_time(t,v,threshold=5):
    """ 
    Threshold here is a dVdt threshold in mV/ms`
    Default threshold is taken from Davidson et al. 2014, PAIN
    """
    assert threshold > 0, 'Rise time threshold is a gradient threshold, should be > 0!'
    dVdt = np.gradient(v,t)
    peak = ap_peak(v)
    peak_idx = peak[1]
    peak_time = t[peak_idx]
    
    # If dVdt is a tuple, second part is gradient
    found_thresholds = []
    for i,gradient in enumerate(dVdt[0:-1]): # Is dVdt a time vector as well?
        if gradient < threshold:
            if dVdt[i+1] > threshold:
                found_thresholds.append(i)
    
    num_threshold = len(found_thresholds)
    if num_threshold == 1:
        threshold_time = t[found_thresholds[0]]
        rise_time = peak_time - threshold_time
        if rise_time < 0:
            #rise_time = 'Rise time < 0: %.3f' % rise_time
            rise_time = np.nan
#        assert rise_time >=0, 'Rise time < 0!'
    elif num_threshold == 0:
        rise_time = np.nan
    elif num_threshold > 1:
#        assert False, 'More than 1 threshold for rise time - APs may not be clearly separated.'
        # Take the first one - later ones are probably rapid spikes e.g. on the shoulder
        threshold_time = t[found_thresholds[0]]
        rise_time = peak_time - threshold_time
    return rise_time
APRiseTime = ap_rise_time # Alias
        
def ap_slope_min_max(t,v):
    dVdt = np.gradient(v,t)
    slope_min = min(dVdt)
    slope_max = max(dVdt)
    ### Need mins and maxes
    return [slope_min,slope_max]
APSlopeMinMax = ap_slope_min_max # Alias    

def ap_width(t, v, alpha, _threshold=5., threshold_type='gradient'):
    """
    Generic ap width calculating function. Alpha determines the fraction of the voltage
    gap between threshold and ap peak that is used to set the voltage threshold.

    Specifically, if Th is the calculate threshold voltage and P is the peak voltage,
    the voltage threshold used depends on alpha so that width threshold WTh = alpha*P + (1-alpha)*Th

    So for full width alpha = 0, as we just use the bare threshold Th, for half width alpha = 0.5,
    and alpha = 1 should give a width of 0 as it goes all the way to the peak.
    
    Defaults are consistent with Davidson et al. 2014 (5 mV/ms gradient to find threshold voltage)

    _threshold named to avoid overlapping with threshold function
    """ 
    
    # Calculate AP threshold and AP peak voltages
    if threshold_type == 'gradient':
        v_threshold = threshold(t, v, _threshold)
    elif threshold_type == 'voltage':
        v_threshold = _threshold
    else:
        raise ValueError("threshold type: {} not recognised".format(threshold_type))
    if np.isnan(v_threshold):
        return np.nan
    
    v_peak = ap_peak(v)[0]
    width_v_threshold = alpha * v_peak + (1.0 - alpha) * v_threshold

    # Find crossing points
    ups, downs = find_threshold_crossings(v, width_v_threshold)

    # Check we have crossings
    if ups and downs:
        last_down = downs[-1]
        first_up = ups[0]
        width = t[last_down] - t[first_up]
        return width
    else:
        return np.nan

def ap_full_width(t,v ,_threshold=5., threshold_type='gradient'):
    """
    Calculate full width of AP by one of two methods, a voltage threshold
    or a voltage/time gradient threshold
    Defaults are consistent with Davidson et al. 2014 (5 mV/ms gradient to find threshold voltage)

    _threshold named to avoid overlapping with threshold function
    """
    
    if threshold_type == 'voltage':
        assert not np.isnan(_threshold), "Threshold {} is nan".format(_threshold)
        ups, downs = find_threshold_crossings(v, _threshold)
    elif threshold_type == 'gradient':
        # Find voltage at which we cross the dvdt threshold
        dvdt = np.gradient(v,t)
        gradient_threshold = None
        for i, _ in enumerate(dvdt[:-1]):
            if (dvdt[i] < _threshold) and (dvdt[i+1] >= _threshold):
                gradient_threshold = v[i]
                break
        # Return if we don't cross the threshold
        if gradient_threshold:
            ups, downs = find_threshold_crossings(v, gradient_threshold)
        else:
            return np.nan
    else:
        raise ValueError("threshold type: {} not recognised".format(threshold_type))

    #print(arr)
    #print(ups,downs)
    num_ups = len(ups)
    num_downs = len(downs)

    if (num_ups < 1) | (num_downs < 1):
        # Not enough crossings
        full_width = np.nan
    elif (num_ups == 1) & (num_downs == 1): 
        # One crossing of threshold each way
        full_width = t[downs[0]] - t[ups[0]]
    elif (num_ups > 1) | (num_downs > 1):
        # Too many crossings
        # Find earliest crossing from below and latest crossing from above
        # to calculate full width
        first_up = ups[0]
        last_down = downs[-1]
        full_width = t[last_down] - t[first_up]

    return full_width

APFullWidth = ap_full_width # Alias


def ap_half_width(t,v, dvdt_threshold=5.):
    """
    Definition from neuroelectro.org:
    AP duration at membrane voltage halfway between AP threshold and AP peak.
    Currently only uses gradient method for finding threshold for simplicity.
    """
    # Calculate AP threshold and AP peak voltages
    v_threshold = threshold(t,v, dvdt_threshold=dvdt_threshold,)
    v_peak = ap_peak(v)[0]
    half_width_v_threshold = (v_threshold + v_peak)/2.

    # Find crossing points
    ups, downs = find_threshold_crossings(v,half_width_v_threshold)

    # Check we have crossings
    if ups and downs:
        last_down = downs[-1]
        first_up = ups[0]
        half_width = t[last_down] - t[first_up]
        return half_width
    else:
        return np.nan

def fit_afterhyperpolarization(traces, dvdt_threshold, max_time_from_peak=50., ahp_model = 'single_exp', full_output=False):
    """ 
    Gather afterhyperpolarisation regions from a set of traces and fit them to a model 
    of a single exponential (other models can be added as needed)
    
    Outputs:
    Returns either amp,tau or if full_output is selected returns five outputs.
    """

    # Fit to single exponential model
    if ahp_model == 'single_exp':
        def model(x, a, b, c):
            return a - b * np.exp(-x/c)
    else:
        raise ValueError('Model \"{}\" not valid'.format(ahp_model))

    # Return function if we have a result we can't fit a hyperpolarisation to
    def hyperpolarisation_fit_failure(full_output):
        if full_output:
            return np.nan, np.nan, np.nan, np.nan,np.nan, np.nan
        else:
            return np.nan, np.nan, np.nan
    
    # Arrange data to contain each interval between peaks (num APs > 1) or peak to end of trace (n=1) 
    """
    This caused a bug for thr last AP since when the stimulus turns off there is a bigger AHP than the AHP itself.
    Instead, for the last trace in the sequence only, check only for 50 ms after the peak. Could also check for some
    multiple of AP full width, but 50 ms should be sufficient for all but very long abnormal APs. 
    """
    num_APs = traces['numAPs']
    if num_APs < 1:
        return hyperpolarisation_fit_failure(full_output) 
    elif num_APs == 1:
        _t = traces['t'][0]
        _v = traces['v'][0]
        max_idx = np.argmax(_v)
        # Check that the peak is not right at the end of the trace
        if max_idx == len(_v)-1:
            return hyperpolarisation_fit_failure(full_output)
        # Get the period between t(peak) and t(peak) + max_time_from_peak
        t_peak = _t[max_idx]
        t_end = t_peak + max_time_from_peak
        end_idx = np.argmin(abs(_t - t_end))

        ts = [_t[max_idx+1:end_idx]] # Single element lists from just after peak to end of max_time_from_peak period
        vs = [_v[max_idx+1:end_idx]] 
    elif num_APs > 1:
        ts = []
        vs = []
        # Traces 1 to N-1
        for i in range(num_APs-1):
            _ts = [traces['t'][idx] for idx in [i, i+1]]
            _vs = [traces['v'][idx] for idx in [i, i+1]]
            max_idxs = [np.argmax(_v) for _v in _vs]
            # Concatenate the two parts of the interval from each trace
            _t_start = _ts[0][max_idxs[0]:]
            _t_end =  _ts[1][:max_idxs[1]-1]
            _v_start =_vs[0][max_idxs[0]:]
            _v_end =  _vs[1][:max_idxs[1]-1] 
            
            _t = np.concatenate([_t_start, _t_end], axis=0)
            _v = np.concatenate([_v_start, _v_end], axis=0)
            ts.append(_t)
            vs.append(_v)

        # Trace N - final trace - use same process as for when there is only 1 AP except change the index
        _t = traces['t'][num_APs-1]
        _v = traces['v'][num_APs-1]
        max_idx = np.argmax(_v)
        # Check that the peak is not right at the end of the trace
        if max_idx == len(_v)-1:
            return hyperpolarisation_fit_failure(full_output)
        # Get the period between t(peak) and t(peak) + max_time_from_peak
        t_peak = _t[max_idx]
        t_end = t_peak + max_time_from_peak
        end_idx = np.argmin(abs(_t - t_end))

        ts.append(_t[max_idx+1:end_idx])
        vs.append(_v[max_idx+1:end_idx])
        
    # For each interval attempt to fit an AHP
    amps = []
    taus = []
    troughs = []
    if full_output: 
        output_ts = []
        output_vs = []
        popts = []
    
    for i, (t, v) in enumerate(zip(ts, vs)):
        # Start from the minimum, until dvdt exceeds the threshold given as input
        min_idx  = np.argmin(v)
        dvdt = np.gradient(v,t)
        threshold_exceeded = dvdt > dvdt_threshold
        if any(threshold_exceeded):
            cutoff_idx = np.where(threshold_exceeded)[0][0] - 1
        else: # Use the whole trace
            cutoff_idx = len(t)-1
        
        t = t[min_idx:cutoff_idx]
        v = v[min_idx:cutoff_idx]
        
        # Check v and t are all there
        if any(np.isnan(v)) | any(np.isnan(t)):
            return hyperpolarisation_fit_failure(full_output)
        
        # If the membrane potential slowly monotonically decreases after a spike, then the min_idx will be the
        # last element of the trace, and so t and v will be empty.

        # Also, if the AHP part of the trace has a very small number of elements, then the calculation might not work.
        # So we will impose a minimum length threshold for the ahp. 
        # With dt = 0.05 ms after downsampling, and a real AHP tau of order 10 ms, any trace of
        # less than length 100 (5 ms) is probably no good, # and any trace less than length 10 (0.5 ms) 
        # is almost certainly not going to contain a viable AHP. 
        # This does assume the default dt though so we should check it is not larger than about 0.1 ms,
        # which would equate to a 1 ms minimum AHP duration.
        
        length_threshold = 10
        if (len(t) <= length_threshold) | (len(v) <= length_threshold):
            return hyperpolarisation_fit_failure(full_output)    
        dt = np.gradient(t)
        assert np.mean(dt) < 0.1, "dt is large, check length_threshold"

        # We can check for this and return failure if it is the case. To do: think about whether this is the best
        # way to handle the lack of an ahp. There could also be cases where we have an AP with no AHP, 
        # and these might give odd tau and amp readings that should not be averaged. For calibration this is fine, 
        # but for mechanistic investigation it might not be so good. 
        
        # Use scipy.optimise.curvefit to fit curve - another option would be to use the more complex 
        # LMFIT library, but I have no experience with it's advantages over the basic scipy lsq fit function
        if ahp_model == 'single_exp':
            t = t - t[0] # Zero out t as model assumes this
            popt, pcov = optimize.curve_fit(model, t, v) # Do the fitting
            # following neuroelectro: https://neuroelectro.org/ephys_prop/index/
            # AHP amplitude is from threshold voltage to trough
            trough = min(v)
            # Need original AP to calculate 
            thresh = threshold(traces['t'][i], traces['v'][i], dvdt_threshold) 
            ahp_amp = trough - thresh # will be -ve
            ahp_tau = popt[2]
            ahp_trough = trough
            " Alternate approaches to calculation"
            """
            calculation_method = False
            if calculation_method == "Simple":
                ahp_amp = trough
            elif calculation_method == "ToRecovery":
                ahp_amp = None
            """
                


        else:
            raise ValueError('Model \"{}\" not valid'.format(ahp_model))

        amps.append(ahp_amp)
        taus.append(ahp_tau)
        troughs.append(ahp_trough)
        if full_output:
            output_ts.append(t)
            output_vs.append(v)
            popts.append(popt)
    
    # Return non averaged output and cutoff times and voltages if full output requested "
    if full_output == True:
        return amps, taus, troughs, output_ts, output_vs, popts
        # Otherwise just return mean amplitude and time constant of the AHP "
    else:
        amp = np.mean(amps)
        tau = np.mean(taus)
        trough = np.mean(troughs)
        return amp, tau, trough

"""
def expFunc(t, amp, slope, start):
    return amp*(1 - np.exp(-slope*t)+start)    
    
maxIdx = []
maxIdx.append(np.argmax(v)) # Get idx of max(v)
maxIdx.append(np.argmax(v2))# Get idx of max(v2)

workingTime = np.concatenate((t[maxIdx[0]:],t2[:maxIdx[1]+1]),0) ### join t[maxIdx1:] up to t2[1:maxIdx[1]]
workingVoltage = np.concatenate((v[maxIdx[0]:],v2[:maxIdx[1]+1]),0) ### join
   
# AHP amplitude
amp = min(workingVoltage)
ampIdx = np.argmin(workingVoltage)

#    dvdt = VoltageGradient(workingTime[ampIdx:], workingVoltage[ampIdx:])
#    plt.plot(workingTime[ampIdx:-1],dvdt)
#    temp = np.argwhere(dvdt > dvdt_threshold) # Temp because we only need the first element
#    takeoffIdx = temp[0][0] # TODO This will break if there's no points above dvdt_threshold
#    plt.plot(workingTime[ampIdx:ampIdx+takeoffIdx],workingVoltage[ampIdx:ampIdx+takeoffIdx])
#    plt.plot(workingTime,workingVoltage)
# AHP time constant
# TO DO!!
# Look up curve fitting    

tau = 'Time constant not implemented'
return amp, tau
"""


def inter_spike_interval(traces):
    """ Calculate average interspike interval from a divided set of traces
        Total interspike interval is the time difference between the first and last peak of a trace,
        divided by the number of intervals (number of APs - 1)
    """
    numAPs = traces['numAPs']
    if numAPs < 2:
        #print('ISI cannot be calculated with < 2 APs')
        return np.nan
    else:
        # Find the peak of the first and last trace
        voltages = traces['v']
        first_spike = np.argmax(voltages[0])
        last_spike = np.argmax(voltages[-1])

        # Get the time difference
        times = traces['t']        
        time_diff = times[-1][last_spike] - times[0][first_spike]
        assert time_diff > 0, 'time_diff for ISI < 0: {}'.format(time_diff)
        # Divide by number of intervals (numAPs - 1) to get mean ISI
        inter_spike_interval = time_diff/(numAPs-1)
        return inter_spike_interval

def absmax(i):
    """
    Returns the largest absolute value present in an array in its raw form
    (e.g. in [-2, 0, 1] it returns -2, in [-2,0,3] it returns 3.)
    """
    # Use the absolute largest value in its raw form
    if max(i) > abs(min(i)):
        return max(i) 
    elif abs(min(i)) >= max(i):
        return min(i)
    else:
        raise ValueError()
        
# ---- Calculating biomarkers over multiple traces ----

def calculate_rmp(traces):
    RMPVals = []
    for i,v in enumerate(traces['v']):
        RMPValue, RMPIdx = RMP(v)
        RMPVals.append(RMPValue)
    return RMPVals
CalculateRMP = calculate_rmp # Alias
    
def calculate_input_res():
    input_res_vals = []
    for i,v in enumerate(traces['v']):
        input_res_vals.append(input_res(v))
    return input_res_vals
CalculateInputRes = calculate_input_res # Alias
    
def calculate_ramp_ap():
    """ 
    Can't remember what this biomarker was supposed to do? 
    We just run ramp simulations and calculate biomarkers on those now.
    """
    # TODO
    return 0
CalculateRampAP = calculate_ramp_ap # Alias
    

def calculate_rheobase(cell_model, amp_step=0.1, amp_max=5., make_plot=False, sim_kwargs=None, search='simple'):
    " Run a series of simulations to calculate rheobase"
    " Rheobase is defined as the threshold current for an infinite duration pulse "
    " We'll try 2 seconds "
     
    # Fill out sim_kwargs with defaults if needed
    if sim_kwargs is None:
        sim_kwargs = {} 
    default_kwargs = {'dur':500., 'delay':1000., 'interval':0., 'num_stims':1, 't_stop':1500.,
            'mechanisms':None, 'make_plot':False, 'plot_type':'default', 'model':cell_model}
    for kwarg in default_kwargs.keys():
        if kwarg in sim_kwargs.keys():
            pass
        else:
            sim_kwargs[kwarg] = default_kwargs[kwarg]

    def rheobase_simulation(amp):
        # Returns simulation amplitude if an AP is found, otherwise returns RHEO_FAIL if no APs found
        sim_kwargs['amp'] = amp
        output = sh.simulation(**sim_kwargs)
        t = output['t']; v = output['v'];
        # Look for an AP, after throwing away the delay period, leave a 1 ms run up to catch the start
        run_up  = 1.
        delay = sim_kwargs['delay']
        stim_period_indices = (t >= (delay-run_up)) # TODO - why did I put tuple here earlier?
        t = t[stim_period_indices]
        v = v[stim_period_indices]        
        traces = split_trace_into_aps(t,v,threshold=0.,time_threshold=5.)
        if traces['numAPs'] > 0: # rheobase found
            if make_plot:
                plot_traces(traces)
            rheobase = amp
            return rheobase
        else:
            return RHEO_FAIL

    amp_min = 0.
    amps = np.arange(amp_min, amp_max, amp_step) # (nA)
    
    # Two search modes 
    # 1. simple starts from amp_min and works up until it finds an AP
    # 2. divide starts from the middle and does a binary search
    # simple should be quicker when rheobase is usually very low and very few models have no rheobase
    # divide should be quicker if rheobase is distributed any other way

    if search == 'simple':
        for amp in amps:
            rheobase = rheobase_simulation(amp)
            if rheobase is not RHEO_FAIL: # Is not is used because np.nan == np.nan reutrns False
                return rheobase
        return RHEO_FAIL

    elif search == 'divide':
        # Divide and conquer algorithm using a binary search
        idx0 = 0
        idxn = len(amps) - 1
        rheobases = np.empty(len(amps))
        rheobases[:] = None

        while idx0 <= idxn:
            midval = (idx0 + idxn)// 2
            rheobase = rheobase_simulations(amps[midval])
            rheobases[midval] = rheobase
            if rheobase is not RHEO_FAIL: # Is not is used because np.nan == np.nan reutrns False
                if midval == 0:
                    # Rheobase is minimum
                    return amps[0]
                elif rheobases[midval-1] is not RHEO_FAIL: # Is not is used because np.nan == np.nan reutrns False
                    # Found minimal amp for an AP - return rheobase
                    return amps[midval]
                else:
                    # AP found but not definitely lowest amp so lower idxn
                    idxn = midval - 1
            elif rheobase is not RHEO_FAIL: # Is not is used because np.nan == np.nan reutrns False
                if midval == (len(amps) - 1):
                    # No rheobase for highest amp
                    return RHEO_FAIL
                elif isinstance(rheobases[midval+1], float):
                    # We've found highest amp with no AP, so one up is rheobase
                    return amps[midval+1]
                else:
                    # No AP found but not definitely highest amp so raise idx0
                    idx0 = midval + 1
            else:
                raise Exception('Rheobase not accepted value.' )
        raise Exception('No rheobase found')
    elif search == 'smart':
        # Simple search but after first two searches upwards we check the max value to check for
        # no rheobase. If the first 5? searches fail we switch to binary.
        # TODO
        pass
CalculateRheobase = calculate_rheobase # Alias for compatibility

def calculate_threshold(traces, dvdt_threshold=5.):
    thresholds = []
    for t,v in zip(traces['t'], traces['v']):
        thresholds.append(threshold(t, v, dvdt_threshold=dvdt_threshold,))
    return thresholds

def calculate_ap_peak(traces):
    ap_peak_vals = []
    for _,v in zip(range(len(traces['t'])),traces['v']):
        ap_peak_vals.append(ap_peak(v)[0])
    return ap_peak_vals
CalculateAPPeak = calculate_ap_peak # Alias
    
def calculate_ap_rise_time(traces,dvdtthreshold=5.):
    ap_rise_time_vals = []
    for t,v in zip(traces['t'],traces['v']):
        ap_rise_time_vals.append(ap_rise_time(t,v,dvdtthreshold))
    return ap_rise_time_vals
CalculateAPRiseTime = calculate_ap_rise_time # Alias
    
def calculate_ap_slope_min_max(traces):
    ap_slope_min_vals = []
    ap_slope_max_vals = [] 
    for t,v in zip(traces['t'],traces['v']):
        dvdt = np.gradient(v,t)
        ap_slope_min_vals.append(min(dvdt))
        ap_slope_max_vals.append(max(dvdt))
    return ap_slope_min_vals, ap_slope_max_vals
CalculateAPSlopeMinMax = calculate_ap_slope_min_max # Alias

def calculate_ap_width(traces, alpha, threshold=0, method='voltage'):
    ap_width_vals = []
    for t,v in zip(traces['t'],traces['v']):
        ap_width_vals.append(ap_width(t,v,alpha,threshold,method))
    return ap_width_vals

def calculate_ap_half_width(traces, threshold=0, method='voltage'):
    alpha = 0.5
    ap_half_width_vals = calculate_ap_width(traces,alpha,threshold,method)
    return ap_half_width_vals

def calculate_ap_full_width(traces,threshold=0, method='voltage'):
    alpha = 0.0 # Calculate at the threshold so set alpha = 0
    ap_full_width_vals = calculate_ap_width(traces,alpha,threshold,method)
    return ap_full_width_vals
CalculateAPFullWidth = calculate_ap_full_width # Alias

def calculate_ahp_amp(traces,dvdt_threshold=5):
    ahp_amp_vals = []
    if traces['numAPs'] > 1:
        for i in range(traces['numAPs']-1):
            t = traces['t'][i]
            v = traces['v'][i]
            t2 = traces['t'][i+1]
            v2 = traces['v'][i+1]
            amp, tau, trough = fit_afterhyperpolarization(t,v,t2,v2,dvdt_threshold)
            AHPAmpVals.append(amp)
    elif traces['numAPs'] == 1:
        v = traces['v'][0]
        max_idx = np.argmax(v)
        working_voltage = v[max_idx:]### join
        amp = min(working_voltage)
        ahp_amp_vals.append(amp)
    return ahp_amp_vals
CalculateAHPAmp = calculate_ahp_amp # Alias
    
def calculate_ahp_tau():
    # TODO
    return 0
CalculateAHPTau = calculate_ahp_tau # Alias

# -- Firing Patterns --
# See Balachandar and Prescott 2018 for algorithms
# TODO: Find algorithms for phasic and burst patterns

def determine_firing_pattern(traces, stim_start, stim_end):
    """
    Define firing pattern of traces as one or more of n types:
    1. Reluctant
    2. Single
    3. Tonic
    4. Delayed
    5. Gap
    6. Phasic - multi-AP firing that ends before end of stimulus
    7. Burst firing
    8. Wide
    9. Repolarisation failure
    """

    def first_spike_delay(traces, stim_start):
        # Find delay between stim start and first spike
        first_spike_v = traces['v'][0]
        first_spike_t = traces['t'][0]
        single_spike_index = ap_peak(first_spike_v)[1]
        single_spike_time = first_spike_t[single_spike_index]
        delay = single_spike_time - stim_start
        #print("delay = {}".format(delay))
        return delay

    def first_two_spikes_isi(traces):
        # Find delay between first and second spikes
        spike_times = []
        for i in [0,1]:
            spike_idx = ap_peak(traces['v'][i])[1]
            spike_times.append(traces['t'][i][spike_idx])
        
        delay = spike_times[1] - spike_times[0]
        return delay

    def second_third_spikes_isi(traces):
        # Find delay between second and third spikes
        spike_times = []
        for i in [1,2]:
            spike_idx = ap_peak(traces['v'][i])[1]
            spike_times.append(traces['t'][i][spike_idx])
        
        delay = spike_times[1] - spike_times[0]
        return delay

    def check_delayed(traces):
        # Check if firing pattern is delayed
        delayed = False
        num_aps = traces['numAPs']
        # Delayed firing pattern criterion for 1 spike:
        # Delay from stim start to first spike is > 100 ms
        if num_aps == 1:
            if first_spike_delay(traces, stim_start) > 100.0:
                delayed = True
        # Delayed firing pattern criterion for  > 1 spike:
        # Delay between stimulus start and firing first spike is > 1.5 
        # times the ISI between spikes 1 and 2.
        elif num_aps > 1:
            if first_spike_delay(traces, stim_start) > 1.5*first_two_spikes_isi(traces):
                delayed  = True
        return delayed

    def check_gap(traces):
        gap = False
        num_aps = traces['numAPs']
        # Gap firing criteria:
        # Number of spikes > 2
        # ISI between spikes 1 and 2 > 1.5 times ISI between spikes 2 and 3
        gap = False
        if num_aps > 2:
            if first_two_spikes_isi(traces) > 1.5*second_third_spikes_isi(traces):
                gap = True
        return gap    


    def check_phasic(traces, stim_end, ratio_threshold=0.25):
        """
        Phasic - firing of multiple APs followed by a period of quiescence. 
        Cases
        1. Idea is use ratio of - Time from last spike to stimulus end:time from first to last spike
        If the ratio is above some threshold.
        2. Simply time from last peak to end of stimulus compared to a threshold.
        """
        
        phasic = False
        # Characterisation cases
        case1 = True
        case2 = False
        
        # First, check we have multiple APs
        # We will class single spikes as single spikes, not phasic.
        num_aps = traces['numAPs']
        if num_aps < 2:
            return False
        
        spike_times = []
        for i in range(num_aps):
            spike_idx = ap_peak(traces['v'][i])[1]
            spike_times.append(traces['t'][i][spike_idx])
        
        # Case 1
        if case1:
            last_spike_to_stim_end = stim_end - spike_times[-1]
            # check stimulus ended before last spike, if not can't be phasic
            if last_spike_to_stim_end > 0:
                first_to_last_spike = spike_times[-1] - spike_times[0]
                assert first_to_last_spike > 0

                ratio = last_spike_to_stim_end/first_to_last_spike
                #print("Ratio = {}".format(ratio))
                if ratio > ratio_threshold:
                    phasic = True
        
        # Case 2
        if case2:
            raw_time_threshold = 50.0
            if last_spike_to_stimulus_end > raw_time_threshold:
                phasic = True

        return phasic

    def check_bursting(traces, stim_start):
        """
        Bursting - bursts of APs separated by rest periods
        Not sure how to characterize currently.
        1. Find all AP peaks.
        2. Divide trace up into quiet periods and firing periods
        Quiet period is region where distance between two APs or last AP and 
        stimulus end is greater than some multiple of the average ISI (median?).
        """
        bursting = False
        return bursting

    def check_wide(traces, mean_width_threshold=10.0):
        """
        Abnormally wide APs - feature seen when inserting hNav 1.8 into mice (Han et al. 2015).
        """
        wide = False

        # Get width of each AP using AP half width biomarker
        # Use half width as we can compare against data in Han et al. 2015
        # Figure 8D shows half-width distributions, which motivated the choice to 
        # set default width threshold for 'wide' designation to 10 ms.
        half_widths = []
        for t, v in zip(traces['t'], traces['v']):
            half_widths.append(ap_half_width(t ,v, dvdt_threshold=5.))

        if half_widths: # Check we have widths
            if np.mean(half_widths) > mean_width_threshold:
                wide = True
        return wide

    def check_rep_fail(traces, rep_fail_threshold=0.0):
        """
        Repolarisation failure - trace does not recover to a reasonably depolarised voltage
        This can be set by user but we'll start with using 0 mV as default threshold and 
        can tune as needed.
        """
        rep_fail = False

        last_trace = traces['v'][-1]
        # Check last element of last trace against threshold
        if last_trace[-1] > rep_fail_threshold:
            rep_fail = True
        return rep_fail


    firing_pattern = []
    num_aps = traces['numAPs']
    if num_aps == 0:
        firing_pattern.append('reluctant')
    elif num_aps == 1:
        firing_pattern.append('single')
        if check_delayed(traces):
            firing_pattern.append('delayed')
        if check_wide(traces):
            firing_pattern.append('wide')
        if check_rep_fail(traces):
            firing_pattern.append('rep_fail')
    elif num_aps > 1:
        firing_pattern.append('multi')
        # Determine if tonic spiking - can't be delayed, gap, phasic or repolarisation failure
        phasic = check_phasic(traces, stim_end, ratio_threshold=0.25)
        delayed = check_delayed(traces)
        gap = check_gap(traces)
        rep_fail = check_rep_fail(traces)
        if (not delayed) and (not gap) and (not phasic) and (not rep_fail):
            firing_pattern.append('tonic')
        if phasic:
            firing_pattern.append('phasic')
        if delayed:
            firing_pattern.append('delayed')
        if gap:
            firing_pattern.append('gap')
        if rep_fail:
            firing_pattern.append('rep_fail')
        # Check wide
        if check_wide(traces, mean_width_threshold=10.0):
            firing_pattern.append('wide')


    #print(" TODO:Bursting")
    return firing_pattern


    
# ---- Plotting ----

def plot_traces(traces):
        for t,v in zip(traces['t'], traces['v']):
            plt.plot(t,v)

# ---- I/O ----

def write_header(biomarker_file):
    string = 'Index'
    for biomarker in db.biomarkerNames:      
        string += (';' + biomarker)
    string += ';' + 'stimAmp'
    string += '\n'
    biomarker_file.write(string)
    return
WriteHeader = write_header # Alias
        
def write_biomarkers(biomarkers,biomarker_file):
    # Write the values of each biomarker in csv format
    string = str(biomarkers['Index'])    
    for biomarker in db.biomarkerNames:        
        string += (';' + str(biomarkers[biomarker]))
    string += (';' + str(biomarkers['stimAmp']))
    string += '\n'
    biomarker_file.write(string)
    return
WriteBiomarkers = write_biomarkers # Alias

# ---- Frequency intensity curve biomarkers ----

class FICurves(object):
    """
    Class to hold FI curve data
    Frequencies
    Amplitudes
    Results
    Which simulations go together
    
    And allow you to extract FI curves, plot them and obtain summary statistics
    """
    
    
    def __init__(self, results, simulations):
        self.results = results.copy()
        self.simulations = simulations.copy()
        self.groups = [] # Data storage for each group
        """ 
        Get all the simulations that have a constant stimulus amplitude and aren't run to rheobase 
        and group by block parameters and stimulus type TODO: And also check all other features (e.g. 
        """
        self.group_simulations()
        self.get_FI_curves()
        
        # Process results to remove parameters
        if 'Parameters' in self.results.columns.levels[0]:
            self.results = self.results.drop('Parameters',axis=1,level=0)
            self.results.columns = self.results.columns.drop('Parameters',level=0)
        
        
        """
        If we need multiple stim unit definitions:
        1. Turn above line into for loop
        2 Create a function called get_stim_amp_designations that gives all the things like nA or pA to search for
        3. Search for any of them in the simulation name and accept those that hit one and only one of them
        """

                
    def group_simulations(self):
        """
        Group simulations together that are the same except for their stimulus amplitudes
        """
        self.groups = []
        for name, sim in self.simulations.items():
            amp, shared_params = self.get_simulation_parameters(sim.protocols) 
            # Check for a fixed amplitude (not a simulation to find rheobase)
            if amp:
                # Check whether there is an existing group with matching shared parameters (scaling factors, stim function)
                group = self.check_for_existing_group(shared_params)
                if group is not None:
                    # Add sim name and amplitude to group as key, val
                    self.groups[group]['simulations'][name] = amp
                    #print("Appending: {}".format(self.groups[group]))
                else:
                    new_group = {'simulations': {name:amp}, 'shared_params':shared_params}
                    #print("Making: {}".format(new_group))
                    self.groups.append(new_group)  
                    
                    
    def get_simulation_parameters(self, sim_protocols):
        """
        Extract needed simulation parameters
        Currently: amplitude, stimulus type and scaling factors
        """          
        amp = sim_protocols['amp']
        shared_params = {}
        shared_params['stim_type'] = sim_protocols['stim_func']
        if 'parameter_scaling' in sim_protocols:
            shared_params['parameter_scaling'] = sim_protocols['parameter_scaling']
        return amp, shared_params
    
    
    def check_for_existing_group(self, shared_params):
        """
        Check groups for a group that mathches other_params.
        Returns:
        Group index if group exists
        None if group does not exist
        """
        # Check if other_params matches all other_params in any other group
        group_idx = None
        for i, group in enumerate(self.groups):
            if shared_params == group['shared_params']:
                assert group_idx is None, "group_idx should equal None, instead: {}".format(group_idx)
                group_idx = i

        return group_idx

                                
    def get_FI_curves(self):
        """
        Use the ISIs to compute firing curves for each group, for models that have non-nan ISIs
        Firing curves are calculated for each simulation group
        """
        num_groups = len(self.groups)
        assert num_groups > 0, "Num groups: {} is not > 0".format(num_groups)
        
        # Iterate through each group
        for group in self.groups:
            """
            Get frequencies for each simulation in the group and build into a dataframe
            Dataframe format:
            rows = model indices
            columns = simulation amplitudes
            """
            idx = self.results.index
            amps = [amp for amp in group['simulations'].values()]
            fi_data = pd.DataFrame(index=idx, columns=amps)
            # Populate this group's fi curve df 
            for sim_name, amp in group['simulations'].items():
                ISIs = self.results.loc[:,(sim_name,'ISI')]
                frequencies = self.calculate_frequencies(ISIs)
                fi_data.loc[:,amp] = frequencies
            
            # Save this group's data
            group['FI'] = fi_data
            
    def plot_FI_curves(self):
        """
        Plot FI curves and maybe compute some summary statistics
        Do scatter and line plots so that if we have a single datapoint for a model it still gets plotted
        """
        num_groups = len(self.groups)
        subplot_dim = int(np.ceil(np.sqrt(num_groups))) # Number of subplots (square)
        plt.figure(figsize=(10,10))
        
        for i, group in enumerate(self.groups):
            plt.subplot(subplot_dim,subplot_dim, i+1)
            fi_data = group['FI'].copy()
            
            for idx in fi_data.index:
                data = fi_data.loc[idx,:]
                plt.plot(data.index, data)
                plt.scatter(data.index, data)
                
            plt.xlim(min(fi_data.columns)-0.1, max(fi_data.columns)+0.1)
            plt.ylim(0, None)
            
            # Make a rough title
            separator = '_'
            for i, sim_name in enumerate(group['simulations']):
                if i == 0:
                    temp_title = sim_name
                else:
                    s1 = temp_title.split(separator)
                    s2 = sim_name.split(separator)
                    title_parts = [part for part in s1 if part in s2]
                    title = separator.join(title_parts)
            plt.title(title)
            plt.xlabel('I (nA)')
            plt.ylabel('f (Hz)')

    def calculate_frequencies(self, ISIs):
        return 1000.0/ISIs # Converts ms to Hz: ISI of 500 ms = 2 Hz, ISI of 2000 ms = 0.5 Hz

        
# ---- Util ----

def get_biomarker_names(biomarker_set='all'):
    ''' 
    Biomarkers TODO from neuroelectro:
    * Input resistance
    * AP Half width
    * Membrane time constant
    * Cell capacitance (fixed by simulation)
    * Maximum firing rate
    * Sag ratio
    * Adaptation ratio
    * First spike latency
    * FI slope
    * Spike rise time
    * Spontaneous firing rate
    * There are others but think they are other names for the same concepts
    '''
    if biomarker_set == 'all':
        biomarker_names = ['Threshold', 'APFullWidth', 'APPeak', 'APRiseTime', 'APSlopeMin', 'APSlopeMax', 'AHPAmp', 'AHPTau', 'AHPTrough', 'ISI', 'RMP', 'Rheobase']
    else:
        raise ValueError('biomarker_set {} not found'.format(biomarker_set))
    return biomarker_names

def find_threshold_crossings(arr, _threshold):
    """
    Find all indices at which a threshold is crossed from above and from below
    in an array. Used for finding indices to compute ap widths and half widths.
    """
    #print("threshold = {}".format(_threshold))
    ups = []
    downs = []
    for i, _ in enumerate(arr[:-1]): # Don't iterate on last element
        # Get crossings of threshold from below
        if arr[i] < _threshold:
            if arr[i+1] >= _threshold:
                ups.append(i)
        # Get crossings of threshold from above
        if arr[i] > _threshold:
            if arr[i+1] <= _threshold:
                downs.append(i)
    return ups, downs


def add_total_width_biomarker(pop, width_biomarker='APHalfWidth', filter_width=False, verbose=False):
    """
    Add a total width biomarker to each simulation in a population's results 
    """
    
    def compute_total_width(df, width_biomarker, filter_width=False):
        """
        Compute the total width of a simulation from its results dataframe
        with optional filtering out of AP Width outliers
        """
        freq = 1000.0/df['ISI']
        numAPs = df['numAPs']
        freq[numAPs == 1] = 1 # Approximation 

        width = df[width_biomarker]
        if filter_width:
            outlier_definition = an.get_outlier_definition(width_biomarker)
            width = width[width < outlier_definition]         
        total_width = width * freq
        total_width = total_width.fillna(0)
        return total_width
    
    simulations = [col for col in pop.results.columns.levels[0] if col not in ['Parameters']]
    for col in simulations:
        if verbose:
            print(col)
        total_width = compute_total_width(df=pop.results[col], 
                                          width_biomarker=width_biomarker,
                                          filter_width=filter_width)
        pop.results.loc[:, (col, 'APTotalWidth')] = total_width




# -*- coding: utf-8 -*-
"""
current_models - library of ionic current models implemented in Python

Created on Mon Apr 10 16:30:04 2017

@author: Oliver Britton
"""

import os
import sys
import pandas as pd
import numpy as np
from matplotlib import pyplot as plt
import seaborn as sns

" Voltage clamp generator functions "


" //--Nav models--\\ "

" -- Nav 1.7 models -- "

def nav17vw(Y,t,voltage_clamp_func,voltage_clamp_params):
    " Human Nav 1.7 from Vasylyev Waxman "
    
    v = voltage_clamp_func(t,voltage_clamp_params)
    
    m = Y[0]
    h = Y[1]
    
    alpha_m = 10.22 - 10.22/(1 + np.exp((v+7.19)/15.43)) # Rate for closed -> open (sort of)
    beta_m = 23.76/(1 + np.exp((v+70.37)/14.53)) # Rate for open->closed
    
    """
    Parameters for kinetics - rate constant (3), 2 voltage shifts, 2 slope coefficients.
    """

    minf = alpha_m/(alpha_m + beta_m)
    mtau = 1/(alpha_m + beta_m)

    alpha_h = 0.0744/(1 + np.exp((v+99.76)/11.07))
    beta_h = 2.54 - 2.54/(1 + np.exp((v+7.8)/10.68))

    hinf = alpha_h/(alpha_h + beta_h)
    htau = 1/(alpha_h + beta_h)
    
    dm = (minf-m)/mtau
    dh = (hinf-h)/htau
    return [dm, dh]

def nav17cw(Y,t,voltage_clamp_func,voltage_clamp_params):
    " Rat? Nav 1.7 from Choi Waxman 2011 "
    v = voltage_clamp_func(t,voltage_clamp_params)
    
    m = Y[0]
    h = Y[1]
    s = Y[2]
    
    alpha_m = 15.5/(1 + np.exp(-(v-5)/(12.08)))
    beta_m = 35.2/(1 + np.exp((v+72.7)/16.7))

    minf = alpha_m/(alpha_m + beta_m)
    mtau = 1/(alpha_m + beta_m)

    alpha_h = 0.38685/(1 + np.exp((v+122.35)/15.29))
    beta_h = -0.00283 + 2.00283/(1 + np.exp(-(v+5.5266)/12.70195)) # Rate is negative if v = -inf?

    hinf = alpha_h/(alpha_h + beta_h)
    htau = 1/(alpha_h + beta_h)

    alpha_s = 0.00003 + 0.00092/(1 + np.exp((v+93.9)/16.6))
    beta_s = 132.05 - 132.05/(1 + np.exp((v-384.9)/28.5))

    sinf = alpha_s/(alpha_s + beta_s)
    stau = 1/(alpha_s + beta_s)
    
    dm = (minf-m)/mtau
    dh = (hinf-h)/htau
    ds = (sinf-s)/stau
    
    return [dm, dh, ds]
    
" -- Nav 1.8 models -- "
def nav18hw(Y,t,voltage_clamp_func,voltage_clamp_params):
    " Human Nav 1.8 from Huang Waxman 20(14?) "
    
    v = voltage_clamp_func(t,voltage_clamp_params)
    
    m = Y[0]
    h = Y[1]
    
    alpha_m = 7.35 - 7.35/(1 + np.exp((v+1.38)/10.9))
    beta_m = 5.97/(1 + np.exp((v+56.43)/18.26))

    minf = alpha_m/(alpha_m + beta_m)
    mtau = 1/(alpha_m + beta_m)

    alpha_h = 0.011 + 1.39/(1 + np.exp((v+78.04)/11.32))
    beta_h = 0.56 - 0.56/(1 + np.exp((v-21.82)/20.03))

    hinf = alpha_h/(alpha_h + beta_h)
    htau = 1/(alpha_h + beta_h)

    dm = (minf-m)/mtau
    dh = (hinf-h)/htau
    
    return [dm, dh]

def nav18tf(Y,t,voltage_clamp_func,voltage_clamp_params):
    " Rat? Nav 1.8 used in Tigerholm model "
    v = voltage_clamp_func(t,voltage_clamp_params)
    
    m = Y[0]
    h = Y[1]
    s = Y[2]
    u = Y[3]
    
    alpha_m = 2.85 - 2.839/(1 + np.exp((v-1.159)/13.95))
    beta_m = 7.6205/(1 + np.exp((v+46.463)/8.8289))

    minf = alpha_m/(alpha_m + beta_m)
    mtau = 1/(alpha_m + beta_m)

    hinf = 1/(1+np.exp((v+32.2)/4))
    htau = 1.218 + 42.043*np.exp(-((v+38.1)**2)/(2*15.19**2))

    alpha_s = 0.001 * 5.4203 / (1 + np.exp((v+79.816)/16.269))
    beta_s  = 0.001 * 5.0757 / (1 + np.exp(-(v+15.968)/11.542))

    sinf = 1/(1+np.exp((v+45.0)/8))
    stau = 1/(alpha_s + beta_s)

    alpha_u =  0.002 * 2.0434 / (1 + np.exp((v+67.499)/19.51))
    beta_u =  0.002 * 1.9952 / (1 + np.exp(-(v+30.963)/14.792))

    uinf = 1/(1+np.exp((v+51.0)/8))
    utau = 1.0/(alpha_u + beta_u) 
    
    dm = (minf-m)/mtau
    dh = (hinf-h)/htau
    ds = (sinf-s)/stau
    du = (uinf-u)/utau
    
    return [dm, dh, ds, du]
    
def nav18cw(Y,t,voltage_clamp_func,voltage_clamp_params):
    " Nav 1.8 model used in Choi Waxman 2011 "
    v = voltage_clamp_func(t,voltage_clamp_params)
    
    m = Y[0]
    h = Y[1]
    
    alpha_m = 2.85 - 2.839/(1 + np.exp((v-1.159)/13.95))
    beta_m = 7.6205/(1 + np.exp((v+46.463)/8.8289))

    minf = alpha_m/(alpha_m + beta_m)
    mtau = 1/(alpha_m + beta_m)

    hinf = 1/(1+np.exp((v+32.2)/4))
    htau = 1.218 + 42.043*np.exp(-((v+38.1)**2)/(2*15.19**2))
    
    dm = (minf-m)/mtau
    dh = (hinf-h)/htau
    
    return [dm, dh]
    
" -- Nav 1.9 models -- "

def nav19hw(Y,t,voltage_clamp_func,voltage_clamp_params):
    " Nav 1.9 model from Huang Waxman 2014"
    m = Y[0]
    h = Y[1]
    s = Y[2]
    
    v = voltage_clamp_func(t,voltage_clamp_params)
    
    alpha_m = 0.751/(1 + np.exp(-(v+32.26)/13.71))
    beta_m = 5.68/(1 + np.exp((v+123.71)/13.94))
    minf = alpha_m/(alpha_m + beta_m)
    mtau = 1/(alpha_m + beta_m)

    alpha_h = 0.082/(1 + np.exp((v+113.69)/17.4))
    beta_h = 0.24/(1 + np.exp(-(v-10.1)/17.2))
    hinf = alpha_h/(alpha_h + beta_h)
    htau = 1/(alpha_h + beta_h)
    
    alpha_s = 0.019/(1 + np.exp((v+154.51)/11.46))
    beta_s = 0.000376/(1 + np.exp(-(v+60.92)/15.79))
    sinf = alpha_s/(alpha_s + beta_s)
    stau = 1/(alpha_s + beta_s)
    
    dm = (minf-m)/mtau
    dh = (hinf-h)/htau
    ds = (sinf-s)/stau
        
    return [dm, dh, ds]
    
def nav19md(Y,t,voltage_clamp_func,voltage_clamp_params):
    " Nav 1.9 model from Maingret 2008"
    m = Y[0]
    h = Y[1]
    s = Y[2]
    
    v = voltage_clamp_func(t,voltage_clamp_params)
    
    
    return [dm, dh, ds]
    
def nav16zm(Y,t,voltage_clamp_func,voltage_clamp_params):
    " Nav 1.6 model from Zach Mainen 1994 "
    m = Y[0]
    h = Y[1]
    
    v = voltage_clamp_func(t,voltage_clamp_params)
    
    vhalf = -43.0
    a_m =  0.182*(v-vhalf)/(1-np.exp((vhalf-v)/6.))
    b_m =  0.124*(-v+vhalf)/(1-np.exp((-vhalf+v)/6.))
    
    m_inf = a_m/(a_m + b_m)
    m_tau = 1./(a_m + b_m)
    
    vhalf_ha = -50.0
    vhalf_hb = -75.0
    q_h = 5.0

    vhalf_inf = -72.0
    qinf = 6.2

    rate_ha = 0.0091
    rate_hb = 0.024

    a_h =  rate_ha*(v-vhalf_ha)/(1-np.exp((vhalf_ha-v)/q_h))
    b_h =  rate_hb*(-v+vhalf_hb)/(1-np.exp((-vhalf_hb+v)/q_h))

    h_inf = 1.0/(1.0 + np.exp((v-vhalf_inf)/qinf))
    h_tau = 1./(a_h + b_h)
    
    dm = (m_inf-m)/m_tau
    dh = (h_inf-h)/h_tau
    
    return [dm, dh]

" Kv models "

def kdr_tf(Y,t,voltage_clamp_func,voltage_clamp_params):
    " Tigerholm version of the Sheets et al. IKdr model "
    " Model was developed from data recorded at 21 oC "
    
    
    v = voltage_clamp_func(t,voltage_clamp_params)
    n = Y[0]
    q10 = 1.0#3.3 # Preserved in case it is useful but disabled
    
    if v > -31.0:
        tau = 0.16+0.8*np.exp(-0.0267*(v+11))
    else:
        tau = 1000*(0.000688 + 1/(np.exp((v+75.2)/6.5) + np.exp(-(v-131.5)/(34.8))))
		
    ninf = 1/(1 + np.exp(-(v+45)/15.4))
    ntau = tau/q10
    
    dn = (ninf-n)/ntau
    return [dn]
    
def km_tf(Y,t,voltage_clamp_func,voltage_clamp_params):
    """ Tigerholm version of the IM current. Current is from multiple sources:
    The voltage dependence of steady-state activation forthe KM current is from
    Maingret et al. (2008), which was derived from Passmore 2003. The KM channel activation has a fast and a slow 
    time constant as described by Passmore et al. (2003). To account for the 
    two time constants, weimplemented one fast (nf) and one slow (ns) gate, 
    combined as follows.
    """
    # g = gbar * (0.25*ns + 0.75*nf)
    v = voltage_clamp_func(t,voltage_clamp_params)
    ns = Y[0]
    nf = Y[1]
    q10 = 1.0#3.3 # Preserved in case it is useful but disabled
    
    if v < -60.0:
        nstau = 219.0*q10
    else:
        nstau = 13.0*v + 1000.0*q10
        
    nftau_alpha = 0.00395*np.exp((v+30.0)/40.0)
    nftau_beta = 0.00395*np.exp(-(v+30.0)/20.0)*q10
    nftau = 1.0/(nftau_alpha + nftau_beta)
    
    ninf = 1.0/(1.0 + np.exp(-(v+30.0)/6.0)) # Threshold is around -30 mV
    
    dns = (ninf-ns)/nstau
    dnf = (ninf-nf)/nftau
    
    return [dns,dnf]
    
def ka_tf(Y,t,voltage_clamp_func,voltage_clamp_params):
    """ Tigerholm version of IA.
    """
    # g = gbar * n * h
    v = voltage_clamp_func(t,voltage_clamp_params)
    n = Y[0]
    h = Y[1]
    q10 = 1.0#3.3 # Preserved in case it is useful but disabled
    
    ninf = (1.0/(1.0 + np.exp(-(v+5.4+15)/16.4)))**4
    ntau = 0.25 + 10.04*np.exp((-(v+24.67)**2)/(2*34.8**2))*q10
		
    hinf = 1.0/(1.0 + np.exp((v+49.9 + 15.0)/4.6))
    htau = 20.0 + 50.0 * np.exp((-(v+40.0)**2)/(2.0*40.0**2))*q10
    
    # Trap for htau following Sheets /ChoiWaxman/Tigerholm - set it to 5 ms if less than 5 ms
    if htau < 5.0:
        htau = 5.0

    dn = (ninf-n)/ntau
    dh = (hinf-h)/htau
    
    return [dn,dh]

""" 
Ca models 

Implemented:
cal_ja - Jaffe et al. 1994 ICaL model. 
can_mi - Model of N-type Ca current from Migliore 95

To do:
SK
BK
Ca diffusion


"""

def cal_ja(Y,t,voltage_clamp_func,voltage_clamp_params):
    """
    Jaffe et al. 1994 ICaL model. 
    """
    v = voltage_clamp_func(t,voltage_clamp_params)
    m = Y[0]
    
    tfa = 1.
    ki = 0.001 # (mM)
    
    cao = 2.5 # Davidson (mM)
    " To do: make cai variable as an input like voltage "
    cai = 1.e-4 # (mM) Roughly values (100 nM) from Intracellular calcium regulation among subpopulations of rat dorsal root ganglion neurons by Lu, Zhang, Gold 2007
    
    celsius = 37.
    
    def alpha(v):
        return 15.69*(81.5 - v)/(np.exp((-1.0*v+81.5)/10.0)-1.0)
    def beta(v):
        return 0.29*np.exp(-v/10.86)
    def KTF(celsius):
        return ((25./293.15)*(celsius + 273.15))
    def efun(z):
        return np.array([1 - i/2 if i < 1e-4 else i/(np.exp(i)-1) for i in z])
    def calc_ghk(v, cai, cao):       
        f = KTF(celsius)/2
        nu = v/f
        return -f*(1. - (cai/cao)*np.exp(nu))*efun(nu)

    a = alpha(v)
    b = beta(v)
    tau = 1./(tfa*(a + b))
    minf = a/(a+b)
    dm = (minf - m)/tau
    
    """ Calculating the current 
    # h gate
    h2 = ki/(ki+cai)
    gcalbar = 0.003
    ghk = calc_ghk(v,cai,cao)
    ical = gcalbar*m*m*h2*ghk
    """
    return [dm]

def can_mi():
    """
    Model of N-type Ca current from Migliore 95
    """
    pass
    
" HCN models "
def hcn_kn(Y,t,voltage_clamp_func,voltage_clamp_params):
    """ 
    Kouranova Ih model with non-specific current (reversal potential should be set at -30 mV 
    """

    v = voltage_clamp_func(t,voltage_clamp_params)
    n_s = Y[0]
    n_f = Y[1]

    ninf_s = 1/(1 + np.exp((v+87.2)/9.7))
    ninf_f = ninf_s

    if v > -70.0:
        tau_ns = 300.0 + 542.0 * np.exp((v+25.0)/20.0)
        tau_nf = 140.0 + 50.0 * np.exp(-(v+25.0)/20.0)
    else:
        tau_ns = 2500.0 + 100.0 * np.exp((v+240.0)/50.0)
        tau_nf = 250.0 + 12.0 * np.exp((v+240.0)/50.0)

    dns = (ninf_s - n_s)/tau_ns
    dnf = (ninf_f - n_f)/tau_nf

    return [dns, dnf]
    
def hcn_tf(Y,t,voltage_clamp_func,voltage_clamp_params):
    """
    Tigerholm version of the Kouranova Ih model which is identical except
    that when you calculate the current you don't use a nonspecific reversal potential and instead split the current between Na+ and K+, 50/50.    
    """
    
    v = voltage_clamp_func(t,voltage_clamp_params)
    n_s = Y[0]
    n_f = Y[1]
    
    ninf_s = 1/(1 + np.exp((v+87.2)/9.7))
    ninf_f = ninf_s

    if v > -70.0:
        tau_ns = 300.0 + 542.0 * np.exp((v+25.0)/20.0)
        tau_nf = 140.0 + 50.0 * np.exp(-(v+25.0)/20.0)
    else:
        tau_ns = 2500.0 + 100.0 * np.exp((v+240.0)/50.0)
        tau_nf = 250.0 + 12.0 * np.exp((v+240.0)/50.0)

    dns = (ninf_s - n_s)/tau_ns
    dnf = (ninf_f - n_f)/tau_nf
    
    return [dns, dnf]

"""
    # ena, ek, + or -?
    Ih_na = 0.5 * g_h (0.5*n_s + 0.5*n_f) * (Vm + ena)
    Ih_k = 0.5 * g_h * (0.5*n_s + 0.5*n_f) * (Vm + ek) 

"""

" Test models "
def nav17test(Y,t,voltage_clamp_func,voltage_clamp_params):
    " Human Nav 1.7 from Vasylyev Waxman "
    
    v = voltage_clamp_func(t,voltage_clamp_params)
    
    m = Y[0]
    h = Y[1]
    
    alpha_m = 10.22 - 10.22/(1 + np.exp((v+7.19)/15.43)) # Rate for closed -> open (sort of)
    beta_m = 23.76/(1 + np.exp((v+70.37)/14.53)) # Rate for open->closed
    
    """
    Parameters for kinetics - rate constant (3), 2 voltage shifts, 2 slope coefficients.
    """

    minf = alpha_m/(alpha_m + beta_m)
    mtau = 1/(alpha_m + beta_m)

    alpha_h = 0.0744/(1 + np.exp((v+99.76)/11.07))
    beta_h = 2.54 - 2.54/(1 + np.exp((v+7.8)/10.68))

    hinf = alpha_h/(alpha_h + beta_h)
    htau = 1/(alpha_h + beta_h)
    
    dm = (minf-m)/mtau
    dh = (hinf-h)/htau
    return [dm, dh]
    

 



"""
Example configuration file to run a simulation saving the traces of all the ionic currents for each model. Be warned this will use a lot of disk space but is necessary if you want
to analyse the ionic currents and plot currentscapes. Check the comments for this script to see where to make changes to a normal simulation to save currents.
"""

import matplotlib
matplotlib.use('Agg')

import time
import sys
import os
import multiprocessing as mp

import drgpom as drg

# Define number of cpu cores to use
# If not specified default to cpu count - 1
if __name__ == '__main__':
    sys.modules['__main__'].__spec__ = None # Without this repeated runnings of this script in Python 3 fail
    # Set cores to cpu count or specified number
    cores = mp.cpu_count()-1
    if len(sys.argv) >= 2:
        for arg in sys.argv[1:]: 
            if 'cores=' in arg:
                core_str = arg.rsplit('=')[-1]
                cores = int(core_str)
                print("Cores = {}".format(cores))

    start = time.time() # For benchmarking
    
    " //-- Main Program begins here --// "

    " --- Load existing population of models --- "

    pop_filename = os.path.join('data', 'example_population.pkl')
    pop = drg.load(pop_filename)
    name = 'example_save_all_ionic_currents'
    pop.name = name
    save_filename = '{}.pkl'.format(name)

    # Simulation parameters
    save_dir = None
    benchmark = True
    rerun = False

    " ---// Setup for recording all ionic currents \\--- "
    save_type = 'both' # Save both trace data files and figures - we need to save the trace data files to save the ionic current data
    outputs = list(pop.mechanisms.keys()) # This tells the simulator to save all "mechanisms" as outputs (mechanisms are what NEURON calls ionic current models).
    print(f"Outputs are: {outputs}") # Prints out which currents we are saving.
    " -------------------------------------------------- "

    " --- Run Simulations --- "

    " --- Simulation 1 - Example Ramp Stimulus --- "
    sim_name = 'example_ramp'
    sim_type = 'iclamp'
    protocols = {
         'amp': None,
         'celsius': 32.0,
         'delay': 500,
         'dur': 500,
         'interval': 0,
         'ions': ['Na','K'],
         'num_stims': 1,
         'outputs': outputs, # This line is vital - here we pass the list of ionic currents to save to the simulation.
         'sampling_freq': 20000,
         'stim_func': 'h.IRamp',
         't_stop': 1500.,
         'v_init': -65.,
         }

    pop.run_simulation(name=sim_name, 
                   simulation_type=sim_type, 
                   protocols=protocols,
                   cores=cores, 
                   save_type=save_type, 
                   save_dir=save_dir, 
                   benchmark=benchmark, 
                   rerun=rerun)


    " --- Simulation 2 - Example Step Stimulus --- "
    sim_name = 'example_step' 
    sim_type = 'iclamp'
    protocols = {
         'amp': None,
         'celsius': 32.0,
         'delay': 500,
         'dur': 800,
         'interval': 0,
         'ions': ['Na','K'],
         'num_stims': 1,
         'outputs': outputs, # This line is vital - here we pass the list of ionic currents to save to the simulation.
         'sampling_freq': 20000,
         'stim_func': 'h.IClamp',
         't_stop': 1500.,
         'v_init': -65.,
         'flags': {'ramp_threshold_sim_for_width':'ramp'} # Use the threshold from the ramp simulation to compute ap full width
         }

    pop.run_simulation(name=sim_name, 
                   simulation_type=sim_type, 
                   protocols=protocols,
                   cores=cores, 
                   save_type=save_type, 
                   save_dir=save_dir, 
                   benchmark=benchmark, 
                   rerun=False)

    " --- Save Results --- "

    print(pop.results.head())
    print("Time taken on {} cores = {}s.".format(cores,time.time()-start))
    pop.save(save_filename)
    print("Current population saved to: {}".format(save_filename))
